import fcntl
import socket

from baseplate.lib.config import EndpointConfiguration


def bind_socket(endpoint: EndpointConfiguration) -> socket.socket:
    sock = socket.socket(endpoint.family, socket.SOCK_STREAM)

    # configure the socket to be auto-closed if we exec() e.g. on reload
    flags = fcntl.fcntl(sock.fileno(), fcntl.F_GETFD)
    fcntl.fcntl(sock.fileno(), fcntl.F_SETFD, flags | fcntl.FD_CLOEXEC)

    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    # on linux, SO_REUSEPORT is supported for IPv4 and IPv6 but not other
    # families. we prefer it when available because it more evenly spreads load
    # among multiple processes sharing a port. (though it does have some
    # downsides, specifically regarding behaviour during restarts)
    if endpoint.family in (socket.AF_INET, socket.AF_INET6):
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)

    sock.bind(endpoint.address)
    sock.listen(128)
    return sock



"""Application metrics via StatsD.

A client for the application metrics aggregator StatsD_. Metrics sent to
StatsD are aggregated and written to graphite. StatsD is generally used for
whole-system health monitoring and insight into usage patterns.

.. testsetup::

    app_config = {"metrics.endpoint": "", "metrics.namespace": "example"}
    do_something = lambda: ()
    do_something_else = lambda: ()
    do_another_thing = lambda: ()

Basic example usage:

.. testcode::

    from baseplate.lib.metrics import metrics_client_from_config

    client = metrics_client_from_config(app_config)
    client.counter("events.connect").increment()
    client.gauge("workers").replace(4)

    with client.timer("something.todo"):
        do_something()
        do_something_else()

If you have multiple metrics to send, you can batch them up for efficiency:

.. testcode::

    with client.batch() as batch:
        batch.counter("froozles").increment()
        batch.counter("blargs").decrement(delta=3)

        with batch.timer("something"):
            do_another_thing()

and the batch will be sent in as few packets as possible when the `with` block
ends.

.. _StatsD: https://github.com/statsd/statsd

"""
import collections
import errno
import logging
import socket
import time

from types import TracebackType
from typing import Any
from typing import DefaultDict
from typing import Dict
from typing import List
from typing import Optional
from typing import Type

from baseplate.lib import config


logger = logging.getLogger(__name__)


def _metric_join(*nodes: bytes) -> bytes:
    return b".".join(node.strip(b".") for node in nodes if node)


def _format_tags(tags: Optional[Dict[str, Any]]) -> Optional[bytes]:
    if not tags:
        return None

    parts = []
    for key, value in tags.items():
        parts.append(f"{key}={str(value)}")
    return b"," + ",".join(parts).encode()


class TransportError(Exception):
    pass


class MessageTooBigTransportError(TransportError):
    def __init__(self, message_size: int):
        super().__init__(f"could not send: message of {message_size} bytes too large")
        self.message_size = message_size


class Transport:
    def send(self, serialized_metric: bytes) -> None:
        raise NotImplementedError

    def flush(self) -> None:
        raise NotImplementedError


class NullTransport(Transport):
    """A transport which doesn't send messages at all."""

    def __init__(self, log_if_unconfigured: bool):
        self.log_if_unconfigured = log_if_unconfigured

    def send(self, serialized_metric: bytes) -> None:
        if self.log_if_unconfigured:
            for metric_line in serialized_metric.splitlines():
                logger.debug("Would send metric %r", metric_line)

    def flush(self) -> None:
        pass


class RawTransport(Transport):
    """A transport which sends messages on a socket."""

    def __init__(
        self,
        endpoint: config.EndpointConfiguration,
        swallow_network_errors: bool = False,
    ):
        self.swallow_network_errors = swallow_network_errors
        self.socket = socket.socket(endpoint.family, socket.SOCK_DGRAM)
        self.socket.connect(endpoint.address)

    def send(self, serialized_metric: bytes) -> None:
        try:
            self.socket.sendall(serialized_metric)
        except OSError as exc:
            if self.swallow_network_errors:
                logger.exception("Failed to send to metrics collector")
                return
            if exc.errno == errno.EMSGSIZE:
                raise MessageTooBigTransportError(len(serialized_metric))
            raise TransportError(exc)

    def flush(self) -> None:
        pass


class BufferedTransport(Transport):
    """A transport which wraps another transport and buffers before sending."""

    def __init__(self, transport: Transport):
        self.transport = transport
        self.buffer: List[bytes] = []

    def send(self, serialized_metric: bytes) -> None:
        self.buffer.append(serialized_metric)

    def flush(self) -> None:
        if self.buffer:
            metrics, self.buffer = self.buffer, []
            message = b"\n".join(metrics)
            self.transport.send(message)


class BaseClient:
    def __init__(self, transport: Transport, namespace: str):
        self.transport = transport
        self.base_tags: Dict[str, Any] = {}
        self.namespace = namespace.encode("ascii")

    def timer(self, name: str, tags: Optional[Dict[str, Any]] = None) -> "Timer":
        """Return a Timer with the given name.

        :param name: The name the timer should have.

        """
        timer_name = _metric_join(self.namespace, name.encode("ascii"))
        return Timer(self.transport, timer_name, {**self.base_tags, **(tags or {})})

    def counter(self, name: str, tags: Optional[Dict[str, Any]] = None) -> "Counter":
        """Return a Counter with the given name.

        The sample rate is currently up to your application to enforce.

        :param name: The name the counter should have.

        """
        counter_name = _metric_join(self.namespace, name.encode("ascii"))
        return Counter(self.transport, counter_name, {**self.base_tags, **(tags or {})})

    def gauge(self, name: str, tags: Optional[Dict[str, Any]] = None) -> "Gauge":
        """Return a Gauge with the given name.

        :param name: The name the gauge should have.

        """
        gauge_name = _metric_join(self.namespace, name.encode("ascii"))
        return Gauge(self.transport, gauge_name, {**self.base_tags, **(tags or {})})

    def histogram(self, name: str, tags: Optional[Dict[str, Any]] = None) -> "Histogram":
        """Return a Histogram with the given name.

        :param name: The name the histogram should have.

        """
        histogram_name = _metric_join(self.namespace, name.encode("ascii"))
        return Histogram(self.transport, histogram_name, {**self.base_tags, **(tags or {})})


class Client(BaseClient):
    """A client for StatsD."""

    def batch(self) -> "Batch":
        """Return a client-like object which batches up metrics.

        Batching metrics can reduce the number of packets that are sent to
        the stats aggregator.

        """
        return Batch(self.transport, self.namespace)


class Batch(BaseClient):
    """A batch of metrics to send to StatsD.

    The batch also supports the `context manager protocol`_, for use with
    Python's ``with`` statement. When the context is exited, the batch will
    automatically :py:meth:`flush`.

    .. _context manager protocol:
        https://docs.python.org/3/reference/datamodel.html#context-managers

    """

    # pylint: disable=super-init-not-called
    def __init__(self, transport: Transport, namespace: bytes):
        self.transport = BufferedTransport(transport)
        self.namespace = namespace
        self.base_tags = {}
        self.counters: Dict[bytes, BatchCounter] = {}

    def __enter__(self) -> "Batch":
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.flush()
        return None  # don't swallow exception

    def flush(self) -> None:
        """Immediately send the batched metrics."""
        counters, self.counters = self.counters, {}
        for counter in counters.values():
            counter.flush()

        try:
            self.transport.flush()
        except MessageTooBigTransportError as exc:
            counters_by_total = list(
                sorted((c for c in counters.values()), key=lambda c: c.total, reverse=True)
            )
            logger.warning(
                "Metrics batch of %d bytes is too large to send, flush more often or reduce "
                "amount done in this request. See https://baseplate.readthedocs.io/en/latest/guide/faq.html#what-do-i-do-about-metrics-batch-of-n-bytes-is-too-large-to-send. Top counters: %s",
                exc.message_size,
                ", ".join(f"{c.name.decode()}={c.total:.0f}" for c in counters_by_total[:10]),
            )
        except TransportError as exc:
            logger.warning("Failed to send metrics batch: %s", exc)

    def counter(self, name: str, tags: Optional[Dict[str, Any]] = None) -> "Counter":
        """Return a BatchCounter with the given name.

        The sample rate is currently up to your application to enforce.
        :param name: The name the counter should have.
        """
        counter_name = _metric_join(self.namespace, name.encode("ascii"))
        batch_counter = self.counters.get(counter_name)
        if batch_counter is None:
            batch_counter = BatchCounter(self.transport, counter_name, tags)
            self.counters[counter_name] = batch_counter

        return batch_counter


class Timer:
    """A timer for recording elapsed times.

    The timer also supports the `context manager protocol`_, for use with
    Python's ``with`` statement. When the context is entered the timer will
    :py:meth:`start` and when exited, the timer will automatically
    :py:meth:`stop`.

    .. _context manager protocol:
        https://docs.python.org/3/reference/datamodel.html#context-managers

    """

    def __init__(
        self,
        transport: Transport,
        name: bytes,
        tags: Optional[Dict[str, Any]] = None,
    ):
        self.transport = transport
        self.name = name
        if tags:
            self.tags = tags
        else:
            self.tags = {}

        self.start_time: Optional[float] = None
        self.stopped: bool = False
        self.sample_rate = 1.0

    def start(self, sample_rate: float = 1.0) -> None:
        """Record the current time as the start of the timer."""
        assert not self.start_time, "timer already started"
        assert not self.stopped, "timer already stopped"

        self.sample_rate = sample_rate
        self.start_time = time.time()

    def stop(self) -> None:
        """Stop the timer and record the total elapsed time."""
        assert self.start_time, "timer not started"
        assert not self.stopped, "timer already stopped"

        now = time.time()
        elapsed = now - self.start_time
        self.send(elapsed, self.sample_rate)
        self.stopped = True

    def send(self, elapsed: float, sample_rate: float = 1.0) -> None:
        """Directly send a timer value without having to stop/start.

        This can be useful when the timing was managed elsewhere and we just
        want to report the result.
        :param elapsed: The elapsed time in seconds to report.
        """
        serialized = self.name
        formatted_tags = _format_tags(self.tags)
        if formatted_tags:
            serialized += formatted_tags
        serialized += f":{(elapsed * 1000.0):g}|ms".encode()
        if sample_rate < 1.0:
            sampling_info = f"@{sample_rate:g}".encode()
            serialized = b"|".join([serialized, sampling_info])
        self.transport.send(serialized)

    def update_tags(self, tags: Dict) -> None:
        assert not self.stopped
        self.tags.update(tags)

    def __enter__(self) -> None:
        self.start()

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.stop()
        return None  # don't swallow exception


class Counter:
    """A counter for counting events over time."""

    def __init__(self, transport: Transport, name: bytes, tags: Optional[Dict[str, Any]] = None):
        self.transport = transport
        self.name = name
        self.tags = tags

    def increment(self, delta: float = 1.0, sample_rate: float = 1.0) -> None:
        """Increment the counter.

        :param delta: The amount to increase the counter by.
        :param sample_rate: What rate this counter is sampled at. [0-1].

        """
        self.send(delta, sample_rate)

    def decrement(self, delta: float = 1.0, sample_rate: float = 1.0) -> None:
        """Decrement the counter.

        This is equivalent to :py:meth:`increment` with delta negated.

        """
        self.increment(delta=-delta, sample_rate=sample_rate)

    def send(self, delta: float, sample_rate: float) -> None:
        """Send the counter to the backend.

        :param delta: The amount to increase the counter by.
        :param sample_rate: What rate this counter is sampled at. [0-1].

        """
        serialized = self.name
        formatted_tags = _format_tags(self.tags)
        if formatted_tags:
            serialized += formatted_tags
        serialized += f":{delta:g}".encode() + b"|c"
        if sample_rate < 1.0:
            sampling_info = f"@{sample_rate:g}".encode()
            serialized = b"|".join([serialized, sampling_info])
        self.transport.send(serialized)


class BatchCounter(Counter):
    """Counter implementation that batches multiple increment calls.

    A new entry in the ``packets`` entry is created for each sample rate.  For
    example, if a counter is incremented multiple times with a sample rate of
    1.0, there will be one entry in ``packets``. If that counter is implemented
    again with a sample rate of 0.5, there will be two entries in ``packets``.
    Each packet has an associated delta value.

    Example usage::
        counter = BatchCounter(transport, "counter_name")
        counter.increment()
        do_something_else()
        counter.increment()
        counter.flush()

    The above example results in one packet indicating an increase of two
    should be applied to "counter_name".
    """

    def __init__(self, transport: Transport, name: bytes, tags: Optional[Dict[str, Any]] = None):
        super().__init__(transport, name)
        self.packets: DefaultDict[float, float] = collections.defaultdict(float)
        self.tags = tags

    def increment(self, delta: float = 1.0, sample_rate: float = 1.0) -> None:
        """Increment the counter.

        :param delta: The amount to increase the counter by.
        :param sample_rate: What rate this counter is sampled at. [0-1].

        """
        self.packets[sample_rate] += delta

    def decrement(self, delta: float = 1.0, sample_rate: float = 1.0) -> None:
        """Decrement the counter.

        This is equivalent to :py:meth:`increment` with delta negated.

        """
        self.increment(delta=-delta, sample_rate=sample_rate)

    @property
    def total(self) -> int:
        return sum(self.packets.values())  # type: ignore

    def flush(self) -> None:
        for sample_rate, delta in self.packets.items():
            super().send(delta, sample_rate)


class Histogram:
    """A bucketed distribution of integer values across a specific range.

    Records data value counts across a configurable integer value range
    with configurable buckets of value precision within that range.

    Configuration of each histogram is managed by the backend service,
    not by this interface. This implementation also depends on histograms
    being supported by the StatsD backend. Specifically, the StatsD
    backend must support the :code:`h` key, e.g. :code:`metric_name:320|h`.
    """

    def __init__(
        self,
        transport: Transport,
        name: bytes,
        tags: Optional[Dict[str, Any]] = None,
    ) -> None:
        self.transport = transport
        self.name = name
        self.tags = tags

    def add_sample(self, value: float) -> None:
        """Add a new value to the histogram.

        This records a new value to the histogram; the bucket it goes in
        is determined by the backend service configurations.
        """
        formatted_tags = _format_tags(self.tags)
        if formatted_tags:
            serialized = self.name + formatted_tags + (f":{value:g}|h".encode())
        else:
            serialized = self.name + (f":{value:g}|h".encode())

        self.transport.send(serialized)


class Gauge:
    """A gauge representing an arbitrary value.

    .. note:: The StatsD protocol supports incrementing/decrementing gauges
        from their current value. We do not support that here because this
        feature is unpredictable in face of the StatsD server restarting and
        the "current value" being lost.

    """

    def __init__(
        self,
        transport: Transport,
        name: bytes,
        tags: Optional[Dict[str, Any]] = None,
    ):
        self.transport = transport
        self.name = name
        self.tags = tags

    def replace(self, new_value: float) -> None:
        """Replace the value held by the gauge.

        This will replace the value held by the gauge with no concern for its
        previous value.

        .. note:: Due to the way the protocol works, it is not possible to
            replace gauge values with negative numbers.

        :param new_value: The new value to store in the gauge.

        """
        assert new_value >= 0, "gauges cannot be replaced with negative numbers"
        formatted_tags = _format_tags(self.tags)
        if formatted_tags:
            serialized = self.name + formatted_tags + (f":{new_value:g}|g".encode())
        else:
            serialized = self.name + (f":{new_value:g}|g".encode())

        self.transport.send(serialized)


def make_client(
    namespace: str,
    endpoint: config.EndpointConfiguration,
    log_if_unconfigured: bool,
    swallow_network_errors: bool = False,
) -> Client:
    """Return a configured client.

    :param namespace: The root key to prefix all metrics with.
    :param endpoint: The endpoint to send metrics to or :py:data:`None`.  If
        :py:data:`None`, the returned client will discard all metrics.
    :param swallow_network_errors: Swallow (log) network errors during sending
        to metrics collector.
    :return: A configured client.

    .. seealso:: :py:func:`baseplate.lib.metrics.metrics_client_from_config`.

    """
    transport: Transport

    if endpoint:
        transport = RawTransport(endpoint, swallow_network_errors=swallow_network_errors)
    else:
        transport = NullTransport(log_if_unconfigured)
    return Client(transport, namespace)


def metrics_client_from_config(raw_config: config.RawConfig) -> Client:
    """Configure and return a metrics client.

    This expects two configuration options:

    ``metrics.namespace``
        The root key to prefix all metrics in this application with.
    ``metrics.endpoint``
        A ``host:port`` pair, e.g. ``localhost:2014``. If an empty string, a
        client that discards all metrics will be returned.
    `metrics.log_if_unconfigured``
        Whether to log metrics when there is no unconfigured endpoint.
        Defaults to false.
    `metrics.swallow_network_errors``
        When false, network errors during sending to metrics collector will
        cause an exception to be thrown. When true, those exceptions are logged
        and swallowed instead.
        Defaults to false.

    :param raw_config: The application configuration which should have
        settings for the metrics client.
    :return: A configured client.

    """
    cfg = config.parse_config(
        raw_config,
        {
            "metrics": {
                "namespace": config.Optional(config.String, default=""),
                "endpoint": config.Optional(config.Endpoint),
                "log_if_unconfigured": config.Optional(config.Boolean, default=False),
                "swallow_network_errors": config.Optional(config.Boolean, default=False),
            }
        },
    )

    # pylint: disable=maybe-no-member
    return make_client(
        namespace=cfg.metrics.namespace,
        endpoint=cfg.metrics.endpoint,
        log_if_unconfigured=cfg.metrics.log_if_unconfigured,
        swallow_network_errors=cfg.metrics.swallow_network_errors,
    )



import logging
import time

from threading import Event
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Mapping
from typing import NamedTuple
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import TYPE_CHECKING
from typing import Union

from cassandra.auth import PlainTextAuthProvider
from cassandra.cluster import _NOT_SET  # pylint: disable=no-name-in-module
from cassandra.cluster import Cluster  # pylint: disable=no-name-in-module
from cassandra.cluster import ExecutionProfile  # pylint: disable=no-name-in-module
from cassandra.cluster import ResponseFuture  # pylint: disable=no-name-in-module
from cassandra.cluster import Session  # pylint: disable=no-name-in-module
from cassandra.query import BoundStatement  # pylint: disable=no-name-in-module
from cassandra.query import PreparedStatement  # pylint: disable=no-name-in-module
from cassandra.query import SimpleStatement  # pylint: disable=no-name-in-module
from prometheus_client import Counter
from prometheus_client import Gauge
from prometheus_client import Histogram

from baseplate import Span
from baseplate.clients import ContextFactory
from baseplate.lib import config
from baseplate.lib.prometheus_metrics import default_latency_buckets
from baseplate.lib.secrets import SecretsStore


class CassandraPrometheusLabels(NamedTuple):
    cassandra_client_name: str
    cassandra_keyspace: str
    cassandra_query_name: str


REQUEST_TIME = Histogram(
    "cassandra_client_latency_seconds",
    "Time spent executing cassandra query",
    CassandraPrometheusLabels._fields + ("cassandra_success",),
    buckets=default_latency_buckets,
)
REQUEST_ACTIVE = Gauge(
    "cassandra_client_active_requests",
    "Current number of active cassandra queries",
    CassandraPrometheusLabels._fields,
)
REQUEST_TOTAL = Counter(
    "cassandra_client_requests_total",
    "Total number of cassandra queries",
    CassandraPrometheusLabels._fields + ("cassandra_success",),
)

if TYPE_CHECKING:
    import cqlmapper.connection


logger = logging.getLogger(__name__)


def cluster_from_config(
    app_config: config.RawConfig,
    secrets: Optional[SecretsStore] = None,
    prefix: str = "cassandra.",
    execution_profiles: Optional[Dict[str, ExecutionProfile]] = None,
    **kwargs: Any,
) -> Cluster:
    """Make a Cluster from a configuration dictionary.

    The keys useful to :py:func:`cluster_from_config` should be prefixed, e.g.
    ``cassandra.contact_points`` etc. The ``prefix`` argument specifies the
    prefix used to filter keys.  Each key is mapped to a corresponding keyword
    argument on the :py:class:`~cassandra.cluster.Cluster` constructor.  Any
    keyword arguments given to this function will be passed through to the
    :py:class:`~cassandra.cluster.Cluster` constructor. Keyword arguments take
    precedence over the configuration file.

    Supported keys:

    * ``contact_points`` (required): comma delimited list of contact points to
      try connecting for cluster discovery
    * ``port``: The server-side port to open connections to.
    * ``credentials_secret`` (optional): the key used to retrieve the database
        credentials from ``secrets`` as a :py:class:`~baseplate.lib.secrets.CredentialSecret`.

    :param execution_profiles: Configured execution profiles to provide to the
        rest of the application.

    """
    assert prefix.endswith(".")
    parser = config.SpecParser(
        {
            "contact_points": config.TupleOf(config.String),
            "port": config.Optional(config.Integer, default=None),
            "credentials_secret": config.Optional(config.String),
        }
    )
    options = parser.parse(prefix[:-1], app_config)

    if options.port:
        kwargs.setdefault("port", options.port)

    if options.credentials_secret:
        if not secrets:
            raise TypeError("'secrets' is required if 'credentials_secret' is set")
        credentials = secrets.get_credentials(options.credentials_secret)
        kwargs.setdefault(
            "auth_provider",
            PlainTextAuthProvider(username=credentials.username, password=credentials.password),
        )

    return Cluster(options.contact_points, execution_profiles=execution_profiles, **kwargs)


class CassandraClient(config.Parser):
    """Configure a Cassandra client.

    This is meant to be used with
    :py:meth:`baseplate.Baseplate.configure_context`.

    See :py:func:`cluster_from_config` for available configuration settings.

    :param keyspace: Which keyspace to set as the default for operations.
    :param client_name: the service-provided name for the client to identify the backends for
        cassandra host. MUST be user specified, MAY be blank if not specified.

    """

    def __init__(self, keyspace: str, client_name: str = "", **kwargs: Any):
        self.keyspace = keyspace
        self.kwargs = kwargs
        self.client_name = client_name

    def parse(self, key_path: str, raw_config: config.RawConfig) -> "CassandraContextFactory":
        cluster = cluster_from_config(raw_config, prefix=f"{key_path}.", **self.kwargs)
        session = cluster.connect(keyspace=self.keyspace)
        return CassandraContextFactory(session, prometheus_client_name=self.client_name)


class CassandraContextFactory(ContextFactory):
    """Cassandra session context factory.

    This factory will attach a proxy object which acts like a
    :py:class:`cassandra.cluster.Session` to an attribute on the
    :py:class:`~baseplate.RequestContext`. The :py:meth:`execute`,
    :py:meth:`execute_async` and :py:meth:`prepare` methods will automatically
    record diagnostic information.

    :param cassandra.cluster.Session session: A configured session object.
    :param prometheus_client_name: the service-provided name for the client to identify the backends
        for cassandra host. MUST be user specified, MAY be blank if not specified.

    """

    def __init__(self, session: Session, prometheus_client_name: Optional[str] = None):
        self.session = session
        self.prepared_statements: Dict[str, PreparedStatement] = {}
        self.prometheus_client_name = prometheus_client_name

    def make_object_for_context(self, name: str, span: Span) -> "CassandraSessionAdapter":
        return CassandraSessionAdapter(
            name,
            span,
            self.session,
            self.prepared_statements,
            prometheus_client_name=self.prometheus_client_name,
        )


class CQLMapperClient(config.Parser):
    """Configure a CQLMapper client.

    This is meant to be used with
    :py:meth:`baseplate.Baseplate.configure_context`.

    See :py:func:`cluster_from_config` for available configuration settings.

    :param keyspace: Which keyspace to set as the default for operations.

    """

    def __init__(self, keyspace: str, **kwargs: Any):
        self.keyspace = keyspace
        self.kwargs = kwargs

    def parse(self, key_path: str, raw_config: config.RawConfig) -> "CQLMapperContextFactory":
        cluster = cluster_from_config(raw_config, prefix=f"{key_path}.", **self.kwargs)
        session = cluster.connect(keyspace=self.keyspace)
        return CQLMapperContextFactory(session)


class CQLMapperContextFactory(CassandraContextFactory):
    """CQLMapper ORM connection context factory.

    This factory will attach a new CQLMapper
    :py:class:`cqlmapper.connection.Connection` to an attribute on the
    :py:class:`~baseplate.RequestContext`. This Connection object will use the
    same proxy object that CassandraContextFactory attaches to a context to run
    queries so the `execute` command will automatically record diagnostic
    information.

    :param cassandra.cluster.Session session: A configured session object.

    """

    def make_object_for_context(self, name: str, span: Span) -> "cqlmapper.connection.Connection":
        # Import inline so you can still use the regular Cassandra integration
        # without installing cqlmapper
        # pylint: disable=redefined-outer-name
        import cqlmapper.connection

        session_adapter = super().make_object_for_context(name, span)
        return cqlmapper.connection.Connection(session_adapter)


def wrap_future(
    response_future: ResponseFuture,
    callback_fn: Callable[..., None],
    callback_args: Any,
    errback_fn: Callable[..., None],
    errback_args: Any,
) -> ResponseFuture:
    """Patch ResponseFuture.result() to wait for callback or errback to complete.

    The callback_fn and errback_fn are given special treatment: they must be
    complete before a result will be returned from ResponseFuture.result().
    They are not given precedence over other callbacks or errbacks, so if
    another callback triggers the response from the service (and the server
    span is closed) the special callback might not complete. The special
    callback is added first and  callbacks are executed in order, so
    generally the special callback should finish before any other callbacks.

    This fixes a race condition where the server span can complete before
    the callback has closed out the child span.

    """
    response_future._callback_event = Event()

    response_future.add_callback(callback_fn, callback_args, response_future._callback_event)
    response_future.add_errback(errback_fn, errback_args, response_future._callback_event)

    def wait_for_callbacks_result(self: ResponseFuture) -> Any:
        exc = None

        try:
            result = ResponseFuture.result(self)
        except Exception as e:
            exc = e

        # wait for either _on_execute_complete or _on_execute_failed to run
        wait_result = self._callback_event.wait(timeout=0.01)
        if not wait_result:
            logger.warning("Cassandra metrics callback took too long. Some metrics may be lost.")

        if exc:
            raise exc  # pylint: disable=E0702

        return result

    # call __get__ to turn wait_for_callbacks_result into a bound method
    bound_method = wait_for_callbacks_result.__get__(  # type: ignore
        response_future, ResponseFuture
    )
    # patch the ResponseFuture instance
    response_future.result = bound_method
    return response_future


class CassandraCallbackArgs(NamedTuple):
    span: Span
    start_time: float
    prom_labels: CassandraPrometheusLabels


def _on_execute_complete(_result: Any, args: CassandraCallbackArgs, event: Event) -> None:
    # TODO: tag with anything from the result set?
    # TODO: tag with any returned warnings
    try:
        args.span.finish()
    finally:
        prom_labels = args.prom_labels._asdict()
        REQUEST_TIME.labels(**prom_labels, cassandra_success="true").observe(
            time.perf_counter() - args.start_time
        )
        REQUEST_TOTAL.labels(**prom_labels, cassandra_success="true").inc()
        REQUEST_ACTIVE.labels(**prom_labels).dec()
        event.set()


def _on_execute_failed(exc: BaseException, args: CassandraCallbackArgs, event: Event) -> None:
    try:
        exc_info = (type(exc), exc, None)
        args.span.finish(exc_info=exc_info)
    finally:
        prom_labels = args.prom_labels._asdict()
        REQUEST_TIME.labels(**prom_labels, cassandra_success="false").observe(
            time.perf_counter() - args.start_time
        )
        REQUEST_TOTAL.labels(**prom_labels, cassandra_success="false").inc()
        REQUEST_ACTIVE.labels(**prom_labels).dec()
        event.set()


RowFactory = Callable[[List[str], List[Tuple]], Any]
Query = Union[str, SimpleStatement, PreparedStatement, BoundStatement]
Parameters = Union[Sequence[Any], Mapping[str, Any]]


class CassandraSessionAdapter:
    def __init__(
        self,
        context_name: str,
        server_span: Span,
        session: Session,
        prepared_statements: Dict[str, PreparedStatement],
        prometheus_client_name: Optional[str] = None,
    ):
        self.context_name = context_name
        self.server_span = server_span
        self.session = session
        self.prepared_statements = prepared_statements
        self.prometheus_client_name = prometheus_client_name

    def __getattr__(self, name: str) -> Any:
        return getattr(self.session, name)

    def execute(
        self,
        query: Query,
        parameters: Optional[Parameters] = None,
        timeout: Union[float, object] = _NOT_SET,
        query_name: Optional[str] = None,
        **kwargs: Any,
    ) -> Any:
        return self.execute_async(
            query, parameters=parameters, timeout=timeout, query_name=query_name, **kwargs
        ).result()

    def execute_async(
        self,
        query: Query,
        parameters: Optional[Parameters] = None,
        timeout: Union[float, object] = _NOT_SET,
        query_name: Optional[str] = None,
        **kwargs: Any,
    ) -> ResponseFuture:
        prom_labels = CassandraPrometheusLabels(
            cassandra_client_name=self.prometheus_client_name
            if self.prometheus_client_name is not None
            else self.context_name,
            cassandra_keyspace=self.session.keyspace,
            cassandra_query_name=query_name if query_name is not None else "",
        )

        REQUEST_ACTIVE.labels(**prom_labels._asdict()).inc()
        start_time = time.perf_counter()
        trace_name = f"{self.context_name}.execute"
        span = self.server_span.make_child(trace_name)
        span.start()
        # TODO: include custom payload
        if isinstance(query, str):
            span.set_tag("statement", query)
        elif isinstance(query, (SimpleStatement, PreparedStatement)):
            span.set_tag("statement", query.query_string)
        elif isinstance(query, BoundStatement):
            span.set_tag("statement", query.prepared_statement.query_string)
        future = self.session.execute_async(query, parameters=parameters, timeout=timeout, **kwargs)
        callback_args = CassandraCallbackArgs(
            span=span,
            start_time=start_time,
            prom_labels=prom_labels,
        )
        future = wrap_future(
            response_future=future,
            callback_fn=_on_execute_complete,
            callback_args=callback_args,
            errback_fn=_on_execute_failed,
            errback_args=callback_args,
        )
        return future

    def prepare(self, query: str, cache: bool = True) -> PreparedStatement:
        """Prepare a CQL statement.

        :param cache: If set to True (default), prepared statements will be
        automatically cached and reused. The cache is keyed on the text of the
        statement. Set to False if you don't want your prepared statements
        cached, which might be advisable if you have a very high-cardinality
        query set. Prepared statements are cached indefinitely, so be wary of
        memory usage.
        """
        if cache:
            try:
                return self.prepared_statements[query]
            except KeyError:
                pass

        trace_name = f"{self.context_name}.prepare"
        with self.server_span.make_child(trace_name) as span:
            span.set_tag("statement", query)
            prepared = self.session.prepare(query)
            if cache:
                self.prepared_statements[query] = prepared
            return prepared



import unittest

from unittest import mock

from baseplate import Baseplate
from baseplate.observers.tracing import make_client
from baseplate.observers.tracing import NullRecorder
from baseplate.observers.tracing import TraceBaseplateObserver
from baseplate.observers.tracing import TraceLocalSpanObserver
from baseplate.observers.tracing import TraceServerSpanObserver

try:
    import webtest

    from pyramid.config import Configurator

    from baseplate.frameworks.pyramid import BaseplateConfigurator
    from baseplate.frameworks.pyramid import StaticTrustHandler
except ImportError:
    raise unittest.SkipTest("pyramid/webtest is not installed")


class TestException(Exception):
    pass


def example_application(request):
    if "error" in request.params:
        raise TestException("this is a test")
    return {"test": "success"}


def local_parent_trace_within_context(request):
    # For testing embedded tracing contexts
    #  See `TracingTests.test_local_tracing_embedded`
    with request.span.make_child("local-req", local=True, component_name="in-context") as span:
        with span.make_child("local-req", local=True, component_name="in-context"):
            pass


class TracingTests(unittest.TestCase):
    def _register_server_mock(self, context, server_span):
        server_span_observer = TraceServerSpanObserver(
            "test-service", "test-hostname", server_span, NullRecorder()
        )
        server_span.register(server_span_observer)
        self.server_span_observer = server_span_observer

    def _register_local_mock(self, span):
        local_span_observer = TraceLocalSpanObserver(
            "test-service", "test-component", "test-hostname", span, NullRecorder()
        )
        self.local_span_ids.append(span.id)
        self.local_span_observers.append(local_span_observer)
        span.register(local_span_observer)

    def setUp(self):
        thread_patch = mock.patch("threading.Thread", autospec=True)
        thread_patch.start()
        self.addCleanup(thread_patch.stop)
        configurator = Configurator()
        configurator.add_route("example", "/example", request_method="GET")
        configurator.add_view(example_application, route_name="example", renderer="json")

        configurator.add_route("local_test", "/local_test", request_method="GET")
        configurator.add_view(
            local_parent_trace_within_context, route_name="local_test", renderer="json"
        )

        self.client = make_client("test-service")
        self.observer = TraceBaseplateObserver(self.client)

        self.baseplate = Baseplate()
        self.baseplate.register(self.observer)

        self.baseplate_configurator = BaseplateConfigurator(
            self.baseplate,
            header_trust_handler=StaticTrustHandler(trust_headers=True),
        )
        configurator.include(self.baseplate_configurator.includeme)
        app = configurator.make_wsgi_app()
        self.local_span_ids = []
        self.local_span_observers = []
        self.test_app = webtest.TestApp(app)

    def test_trace_on_inbound_request(self):
        with mock.patch.object(
            TraceBaseplateObserver, "on_server_span_created", side_effect=self._register_server_mock
        ):
            self.test_app.get("/example")
            span = self.server_span_observer._serialize()
            self.assertEqual(span["name"], "example")
            self.assertEqual(len(span["annotations"]), 2)
            self.assertEqual(span["parentId"], 0)

    def test_local_tracing_embedded(self):
        with mock.patch.object(
            TraceBaseplateObserver, "on_server_span_created", side_effect=self._register_server_mock
        ), mock.patch.object(
            TraceServerSpanObserver, "on_child_span_created", side_effect=self._register_local_mock
        ), mock.patch.object(
            TraceLocalSpanObserver, "on_child_span_created", side_effect=self._register_local_mock
        ):

            self.test_app.get("/local_test")
            # Verify that child span can be created within a local span context
            #  and parent IDs are inherited accordingly.
            span = self.local_span_observers[-1]._serialize()
            self.assertEqual(span["name"], "local-req")
            self.assertEqual(len(span["annotations"]), 0)
            self.assertEqual(span["parentId"], self.local_span_ids[-2])



import time

from typing import List
from typing import NamedTuple
from typing import Optional


class SerializedBatch(NamedTuple):
    item_count: int
    serialized: bytes


class BatchFull(Exception):
    pass


class Batch:
    def add(self, item: Optional[bytes]) -> None:
        raise NotImplementedError

    def serialize(self) -> SerializedBatch:
        raise NotImplementedError

    def reset(self) -> None:
        raise NotImplementedError


class RawJSONBatch(Batch):
    def __init__(self, max_size: int):
        self.max_size = max_size
        self.reset()

    def add(self, item: Optional[bytes]) -> None:
        if not item:
            return

        serialized_size = len(item) + 1  # the comma at the end

        if self._size + serialized_size > self.max_size:
            raise BatchFull

        self._items.append(item)
        self._size += serialized_size

    def serialize(self) -> SerializedBatch:
        return SerializedBatch(
            item_count=len(self._items), serialized=b"[" + b",".join(self._items) + b"]"
        )

    def reset(self) -> None:
        self._items: List[bytes] = []
        self._size = 2  # the [] that wrap the json list


class TimeLimitedBatch(Batch):
    def __init__(self, inner: Batch, max_age: float):
        self.batch = inner
        self.batch_start: Optional[float] = None
        self.max_age = max_age

    @property
    def age(self) -> float:
        if not self.batch_start:
            return 0
        return time.time() - self.batch_start

    def add(self, item: Optional[bytes]) -> None:
        if self.age >= self.max_age:
            raise BatchFull

        self.batch.add(item)

        if not self.batch_start:
            self.batch_start = time.time()

    def serialize(self) -> SerializedBatch:
        return self.batch.serialize()

    def reset(self) -> None:
        self.batch.reset()
        self.batch_start = None



import contextlib
import logging
import unittest

from importlib import reload
from unittest import mock

import gevent.monkey
import pytest

from baseplate import Baseplate
from baseplate import BaseplateObserver
from baseplate import ServerSpanObserver
from baseplate import SpanObserver
from baseplate import TraceInfo
from baseplate.clients.thrift import ThriftClient
from baseplate.frameworks.thrift import baseplateify_processor
from baseplate.lib import config
from baseplate.lib.thrift_pool import ThriftConnectionPool
from baseplate.observers.timeout import ServerTimeout
from baseplate.observers.timeout import TimeoutBaseplateObserver
from baseplate.server import make_listener
from baseplate.server.thrift import make_server
from baseplate.thrift import BaseplateService
from baseplate.thrift import BaseplateServiceV2
from baseplate.thrift.ttypes import Error
from baseplate.thrift.ttypes import ErrorCode
from baseplate.thrift.ttypes import IsHealthyProbe
from baseplate.thrift.ttypes import IsHealthyRequest

from . import FakeEdgeContextFactory
from .test_thrift import TestService


@contextlib.contextmanager
def serve_thrift(handler, server_spec, server_span_observer=None, baseplate_observer=None):
    # create baseplate root
    baseplate = Baseplate()

    if server_span_observer:

        class TestBaseplateObserver(BaseplateObserver):
            def on_server_span_created(self, context, server_span):
                server_span.register(server_span_observer)

        baseplate.register(TestBaseplateObserver())

    if baseplate_observer:
        baseplate.register(baseplate_observer)

    # set up the server's processor
    logger = mock.Mock(spec=logging.Logger)
    edge_context_factory = FakeEdgeContextFactory()
    processor = server_spec.Processor(handler)
    processor = baseplateify_processor(
        processor, logger, baseplate, edge_context_factory, convert_to_baseplate_error=True
    )

    # bind a server socket on an available port
    server_bind_endpoint = config.Endpoint("127.0.0.1:0")
    listener = make_listener(server_bind_endpoint)
    server = make_server({"stop_timeout": "1 millisecond"}, listener, processor)

    # figure out what port the server ended up on
    server_address = listener.getsockname()
    server.endpoint = config.Endpoint(f"{server_address[0]}:{server_address[1]}")

    # run the server until our caller is done with it
    server_greenlet = gevent.spawn(server.serve_forever)
    try:
        yield server
    finally:
        server_greenlet.kill()


@contextlib.contextmanager
def raw_thrift_client(endpoint, client_spec):
    pool = ThriftConnectionPool(endpoint)
    with pool.connection() as client_protocol:
        yield client_spec.Client(client_protocol)


@contextlib.contextmanager
def baseplate_thrift_client(
    endpoint,
    client_spec,
    client_span_observer=None,
    timeout=None,
):
    app_config = {
        "baseplate.service_name": "fancy test client",
        "example_service.endpoint": str(endpoint),
    }
    if timeout:
        app_config["example_service.timeout"] = timeout
    baseplate = Baseplate(app_config=app_config)

    if client_span_observer:

        class TestServerSpanObserver(ServerSpanObserver):
            def on_child_span_created(self, span):
                span.register(client_span_observer)

        observer = TestServerSpanObserver()

        class TestBaseplateObserver(BaseplateObserver):
            def on_server_span_created(self, context, span):
                span.register(observer)

        baseplate.register(TestBaseplateObserver())

    context = baseplate.make_context_object()
    trace_info = TraceInfo.from_upstream(
        trace_id="1234", parent_id="2345", span_id="3456", flags=4567, sampled=True
    )

    baseplate.configure_context({"example_service": ThriftClient(client_spec.Client)})

    baseplate.make_server_span(context, "example_service.example", trace_info)

    context.raw_edge_context = FakeEdgeContextFactory.RAW_BYTES

    yield context


class GeventPatchedTestCase(unittest.TestCase):
    def setUp(self):
        gevent.monkey.patch_socket()

    def tearDown(self):
        import socket

        reload(socket)
        gevent.monkey.saved.clear()


class ThriftTraceHeaderTests(GeventPatchedTestCase):
    def test_user_agent(self):
        """We should accept user-agent headers and apply them to the server span tags."""

        class Handler(TestService.Iface):
            def example(self, context):
                return True

        handler = Handler()

        server_span_observer = mock.Mock(spec=ServerSpanObserver)
        with serve_thrift(handler, TestService, server_span_observer) as server:
            with baseplate_thrift_client(server.endpoint, TestService) as context:
                context.example_service.example()

        server_span_observer.on_set_tag.assert_called()

    def test_no_headers(self):
        """We should accept requests without headers and generate a trace."""

        class Handler(TestService.Iface):
            def __init__(self):
                self.server_span = None

            def example(self, context):
                self.server_span = context.span
                return True

        handler = Handler()

        with serve_thrift(handler, TestService) as server:
            with raw_thrift_client(server.endpoint, TestService) as client:
                client_result = client.example()

        self.assertIsNotNone(handler.server_span)
        self.assertGreaterEqual(len(handler.server_span.id), 0)
        self.assertTrue(client_result)

    def test_header_propagation(self):
        """If the client sends headers, we should set the trace up accordingly."""
        trace_id = "1234"
        parent_id = "2345"
        span_id = "3456"
        flags = 4567
        sampled = 1

        class Handler(TestService.Iface):
            def __init__(self):
                self.server_span = None

            def example(self, context):
                self.server_span = context.span
                return True

        handler = Handler()

        with serve_thrift(handler, TestService) as server:
            with raw_thrift_client(server.endpoint, TestService) as client:
                transport = client._oprot.trans
                transport.set_header(b"Trace", trace_id.encode())
                transport.set_header(b"Parent", parent_id.encode())
                transport.set_header(b"Span", span_id.encode())
                transport.set_header(b"Flags", str(flags).encode())
                transport.set_header(b"Sampled", str(sampled).encode())
                client_result = client.example()

        self.assertIsNotNone(handler.server_span)
        self.assertEqual(handler.server_span.trace_id, trace_id)
        self.assertEqual(handler.server_span.parent_id, parent_id)
        self.assertEqual(handler.server_span.id, span_id)
        self.assertEqual(handler.server_span.flags, flags)
        self.assertEqual(handler.server_span.sampled, sampled)
        self.assertTrue(client_result)

    def test_optional_headers_optional(self):
        """Test that we accept traces from clients that don't include all headers."""
        trace_id = "1234"
        parent_id = "2345"
        span_id = "3456"

        class Handler(TestService.Iface):
            def __init__(self):
                self.server_span = None

            def example(self, context):
                self.server_span = context.span
                return True

        handler = Handler()

        with serve_thrift(handler, TestService) as server:
            with raw_thrift_client(server.endpoint, TestService) as client:
                transport = client._oprot.trans
                transport.set_header(b"Trace", trace_id.encode())
                transport.set_header(b"Parent", parent_id.encode())
                transport.set_header(b"Span", span_id.encode())
                client_result = client.example()

        self.assertIsNotNone(handler.server_span)
        self.assertEqual(handler.server_span.trace_id, trace_id)
        self.assertEqual(handler.server_span.parent_id, parent_id)
        self.assertEqual(handler.server_span.id, span_id)
        self.assertEqual(handler.server_span.flags, None)
        self.assertEqual(handler.server_span.sampled, False)
        self.assertTrue(client_result)

    def test_budget_header(self):
        """Test that the budget header is set in the headers if the client sets it."""
        budget = "1234"

        class Handler(TestService.Iface):
            def __init__(self):
                self.server_span = None

            def example(self, context):
                self.server_span = context.span
                self.context = context
                return True

        handler = Handler()

        with serve_thrift(handler, TestService) as server:
            with raw_thrift_client(server.endpoint, TestService) as client:
                transport = client._oprot.trans
                transport.set_header(b"Deadline-Budget", budget.encode())
                client_result = client.example()

        self.assertEqual(handler.context.headers.get(b"Deadline-Budget").decode(), budget)
        self.assertTrue(client_result)


class ThriftEdgeRequestHeaderTests(GeventPatchedTestCase):
    def _test(self, header_name=None):
        class Handler(TestService.Iface):
            def __init__(self):
                self.edge_context = None

            def example(self, context):
                self.edge_context = context.edge_context
                return True

        handler = Handler()

        with serve_thrift(handler, TestService) as server:
            with raw_thrift_client(server.endpoint, TestService) as client:
                if header_name:
                    transport = client._oprot.trans
                    transport.set_header(header_name, FakeEdgeContextFactory.RAW_BYTES)
                client_result = client.example()

        assert client_result is True
        return handler.edge_context

    def test_edge_request_context(self):
        """If the client sends an edge-request header we should parse it."""
        edge_context = self._test(b"Edge-Request")
        assert edge_context == FakeEdgeContextFactory.DECODED_CONTEXT

    def test_edge_request_context_case_insensitive(self):
        edge_context = self._test(b"edge-request")
        assert edge_context == FakeEdgeContextFactory.DECODED_CONTEXT

    def test_no_edge_context(self):
        edge_context = self._test()
        assert edge_context is None


class ThriftServerSpanTests(GeventPatchedTestCase):
    def test_server_span_starts_and_stops(self):
        """The server span should start/stop appropriately."""

        class Handler(TestService.Iface):
            def example(self, context):
                return True

        handler = Handler()

        server_span_observer = mock.Mock(spec=ServerSpanObserver)
        with serve_thrift(handler, TestService, server_span_observer) as server:
            with raw_thrift_client(server.endpoint, TestService) as client:
                client.example()

        server_span_observer.on_start.assert_called_once_with()
        server_span_observer.on_finish.assert_called_once_with(None)

    def test_expected_exception_not_passed_to_server_span_finish(self):
        """If the server returns an expected exception, don't count it as failure."""

        class Handler(TestService.Iface):
            def example(self, context):
                raise TestService.ExpectedException()

        handler = Handler()

        server_span_observer = mock.Mock(spec=ServerSpanObserver)
        with serve_thrift(handler, TestService, server_span_observer) as server:
            with raw_thrift_client(server.endpoint, TestService) as client:
                with self.assertRaises(TestService.ExpectedException):
                    client.example()

        server_span_observer.on_start.assert_called_once_with()
        server_span_observer.on_finish.assert_called_once_with(None)

    def test_unexpected_exception_passed_to_server_span_finish(self):
        """If the server returns an unexpected exception, mark a failure."""

        class UnexpectedException(Exception):
            pass

        class Handler(TestService.Iface):
            def example(self, context):
                raise UnexpectedException

        handler = Handler()

        server_span_observer = mock.Mock(spec=ServerSpanObserver)
        with serve_thrift(handler, TestService, server_span_observer) as server:
            with raw_thrift_client(server.endpoint, TestService) as client:
                with self.assertRaises(Error):
                    client.example()

        server_span_observer.on_start.assert_called_once_with()
        self.assertEqual(server_span_observer.on_finish.call_count, 1)
        _, captured_exc, _ = server_span_observer.on_finish.call_args[0][0]
        self.assertIsInstance(captured_exc, UnexpectedException)


class ThriftClientSpanTests(GeventPatchedTestCase):
    def test_client_span_starts_and_stops(self):
        """The client span should start/stop appropriately."""

        class Handler(TestService.Iface):
            def example(self, context):
                return True

        handler = Handler()

        client_span_observer = mock.Mock(spec=SpanObserver)
        with serve_thrift(handler, TestService) as server:
            with baseplate_thrift_client(
                server.endpoint, TestService, client_span_observer
            ) as context:
                context.example_service.example()

        client_span_observer.on_start.assert_called_once_with()
        client_span_observer.on_finish.assert_called_once_with(None)

    def test_expected_exception_not_passed_to_client_span_finish(self):
        """If the server returns an expected exception, don't count it as failure."""

        class Handler(TestService.Iface):
            def example(self, context):
                raise TestService.ExpectedException()

        handler = Handler()

        client_span_observer = mock.Mock(spec=SpanObserver)
        with serve_thrift(handler, TestService) as server:
            with baseplate_thrift_client(
                server.endpoint, TestService, client_span_observer
            ) as context:
                with self.assertRaises(TestService.ExpectedException):
                    context.example_service.example()

        client_span_observer.on_start.assert_called_once_with()
        client_span_observer.on_finish.assert_called_once_with(None)

    def test_unexpected_exception_passed_to_client_span_finish(self):
        """If the server returns an unexpected exception, mark a failure."""

        class UnexpectedException(Exception):
            pass

        class Handler(TestService.Iface):
            def example(self, context):
                raise UnexpectedException

        handler = Handler()

        client_span_observer = mock.Mock(spec=SpanObserver)
        with serve_thrift(handler, TestService) as server:
            with baseplate_thrift_client(
                server.endpoint, TestService, client_span_observer
            ) as context:
                with self.assertRaises(Error):
                    context.example_service.example()

        client_span_observer.on_start.assert_called_once_with()
        self.assertEqual(client_span_observer.on_finish.call_count, 1)
        _, captured_exc, _ = client_span_observer.on_finish.call_args[0][0]
        self.assertIsInstance(captured_exc, Error)


class ThriftEndToEndTests(GeventPatchedTestCase):
    def test_end_to_end(self):
        class Handler(TestService.Iface):
            def __init__(self):
                self.edge_context = None

            def example(self, context):
                self.edge_context = context.edge_context
                return True

        handler = Handler()

        span_observer = mock.Mock(spec=SpanObserver)
        with serve_thrift(handler, TestService) as server:
            with baseplate_thrift_client(server.endpoint, TestService, span_observer) as context:
                context.example_service.example()

        assert handler.edge_context == FakeEdgeContextFactory.DECODED_CONTEXT

    def test_budget_header_pool_timeout(self):
        """Test that the budget header is set in the headers with the pool timeout."""
        retry_timeout_seconds = 100.0

        class Handler(TestService.Iface):
            def example(self, context):
                self.context = context
                return True

        handler = Handler()

        span_observer = mock.Mock(spec=SpanObserver)
        with serve_thrift(handler, TestService) as server:
            with baseplate_thrift_client(
                server.endpoint, TestService, span_observer, timeout="1 second"
            ) as context:
                with context.example_service.retrying(
                    attempts=3, budget=retry_timeout_seconds
                ) as svc:
                    svc.example()

        # this should be 1 second (1000 ms) for the pool timeout
        self.assertAlmostEqual(handler.context.deadline_budget, 1.0)

    def test_budget_header_retry_timeout(self):
        """Test that the budget header is set in the headers with the retry timeout."""
        retry_timeout_seconds = 0.1

        class Handler(TestService.Iface):
            def example(self, context):
                self.context = context
                return True

        handler = Handler()

        span_observer = mock.Mock(spec=SpanObserver)
        with serve_thrift(handler, TestService) as server:
            with baseplate_thrift_client(
                server.endpoint, TestService, span_observer, timeout="1000 seconds"
            ) as context:
                with context.example_service.retrying(
                    attempts=3, budget=retry_timeout_seconds
                ) as svc:
                    svc.example()

        self.assertAlmostEqual(handler.context.deadline_budget, retry_timeout_seconds)

    def test_budget_header_budget_and_backoff(self):
        """Test that the budget header is set in the headers with the backoff timeout."""
        retry_timeout_seconds = 1.0
        backoff = 1.0

        class Handler(TestService.Iface):
            def example(self, context):
                self.context = context
                return True

        handler = Handler()

        span_observer = mock.Mock(spec=SpanObserver)
        with serve_thrift(handler, TestService) as server:
            with baseplate_thrift_client(
                server.endpoint, TestService, span_observer, timeout="1000 seconds"
            ) as context:
                with context.example_service.retrying(
                    attempts=3, budget=retry_timeout_seconds, backoff=backoff
                ) as svc:
                    svc.example()

        self.assertAlmostEqual(handler.context.deadline_budget, retry_timeout_seconds)

    def test_budget_timeout_from_client(self):
        """Test that the server times out when passed a short timeout from the client."""
        retry_timeout_seconds = 0.25

        class Handler(TestService.Iface):
            def example(self, context):
                self.context = context
                with pytest.raises(ServerTimeout):
                    gevent.sleep(1)
                return True

        handler = Handler()

        span_observer = mock.Mock(spec=SpanObserver)
        timeout_observer = TimeoutBaseplateObserver.from_config(
            {"server_timeout.default": "1 hour"}
        )
        with serve_thrift(handler, TestService, baseplate_observer=timeout_observer) as server:
            with baseplate_thrift_client(
                server.endpoint, TestService, span_observer, timeout="1000 seconds"
            ) as context:
                with context.example_service.retrying(
                    attempts=3, budget=retry_timeout_seconds
                ) as svc:
                    svc.example()

        self.assertAlmostEqual(handler.context.deadline_budget, retry_timeout_seconds)


class ThriftHealthcheck(GeventPatchedTestCase):
    def test_v2_client_v1_server(self):
        class Handler(BaseplateService.Iface):
            def is_healthy(self, context):
                return True

        handler = Handler()

        span_observer = mock.Mock(spec=SpanObserver)
        with serve_thrift(handler, BaseplateService) as server:
            with baseplate_thrift_client(
                server.endpoint, BaseplateServiceV2, span_observer
            ) as context:
                healthy = context.example_service.is_healthy(
                    request=IsHealthyRequest(probe=IsHealthyProbe.READINESS),
                )
                self.assertTrue(healthy)

    def test_v2_client_v2_server(self):
        class Handler(BaseplateServiceV2.Iface):
            def __init__(self):
                self.probe = None

            def is_healthy(self, context, req=None):
                self.probe = req.probe
                return True

        handler = Handler()

        span_observer = mock.Mock(spec=SpanObserver)
        with serve_thrift(handler, BaseplateServiceV2) as server:
            with baseplate_thrift_client(
                server.endpoint, BaseplateServiceV2, span_observer
            ) as context:
                healthy = context.example_service.is_healthy(
                    request=IsHealthyRequest(probe=IsHealthyProbe.LIVENESS),
                )
                self.assertTrue(healthy)
                self.assertEqual(handler.probe, IsHealthyProbe.LIVENESS)


class ThriftErrorReplacementTests(GeventPatchedTestCase):
    def test_server_replaces_unhandled_errors(self):
        """The server span should start/stop appropriately."""

        class Handler(TestService.Iface):
            def example(self, context):
                raise Exception("foo")

        handler = Handler()

        server_span_observer = mock.Mock(spec=ServerSpanObserver)
        with serve_thrift(handler, TestService, server_span_observer) as server:
            with raw_thrift_client(server.endpoint, TestService) as client:
                with self.assertRaises(Error) as exc_info:
                    client.example()
        self.assertEqual(exc_info.exception.code, ErrorCode.INTERNAL_SERVER_ERROR)


class ThriftPrometheusMetricsTests(GeventPatchedTestCase):
    def reset_metrics(self, metrics):
        if not metrics:
            return

        try:
            metrics.get_active_requests_metric().clear()
            metrics.get_latency_seconds_metric().clear()
            metrics.get_requests_total_metric().clear()
        except Exception:
            pass

    def assert_correct_metric(
        self, metric, want_count, want_sample, want_name, want_labels, want_value
    ):
        m = metric.collect()
        self.assertEqual(len(m), 1)
        self.assertEqual(len(m[0].samples), want_count)
        sample = m[0].samples[want_sample]
        got_name = sample[0]
        self.assertEqual(got_name, want_name)
        got_labels = sample[1]
        self.assertEqual(got_labels, want_labels)
        got_value = sample[2]
        self.assertEqual(got_value, want_value)



"""Watch nodes in ZooKeeper and sync their contents to disk on change."""
import argparse
import configparser
import json
import logging
import os
import sys
import time

from enum import Enum
from pathlib import Path
from typing import Any
from typing import NoReturn
from typing import Optional

import boto3  # type: ignore

from botocore import UNSIGNED  # type: ignore
from botocore.client import ClientError  # type: ignore
from botocore.client import Config
from botocore.exceptions import EndpointConnectionError  # type: ignore
from kazoo.client import KazooClient
from kazoo.protocol.states import ZnodeStat

from baseplate.lib import config
from baseplate.lib.live_data.zookeeper import zookeeper_client_from_config
from baseplate.lib.secrets import secrets_store_from_config
from baseplate.server import EnvironmentInterpolation


logger = logging.getLogger(__name__)


HEARTBEAT_INTERVAL = 300


class LoaderException(Exception):
    pass


class LoaderType(Enum):
    PASSTHROUGH = "PASSTHROUGH"
    S3 = "S3"

    @classmethod
    def _missing_(cls, value: object) -> "LoaderType":
        logger.error(
            "Loader Type %s has not been implemented yet. Defaulting to PASSTHROUGH", value
        )
        return cls.PASSTHROUGH


class NodeWatcher:
    def __init__(self, dest: str, owner: int, group: int, mode: int):
        self.dest = dest
        self.owner = owner
        self.group = group
        self.mode = mode

    def handle_empty_data(self) -> None:
        # the data node does not exist
        try:
            logger.info("Removing %r; watched node deleted.", self.dest)
            os.unlink(self.dest)
        except OSError as exc:
            logger.debug("%s: couldn't unlink: %s", self.dest, exc)

    def on_change(self, data: Optional[bytes], _znode_stat: ZnodeStat) -> None:
        if data is None:
            self.handle_empty_data()
            return

        loader_type = _parse_loader_type(data)
        if loader_type == LoaderType.S3:
            try:
                data = _load_from_s3(data)
            except LoaderException:
                logger.error("Failed to load data from S3. Not writing to destination file")
                return

        if data is None:
            logger.warning(
                "No data to write to destination file. Something is likely misconfigured."
            )
            return

        # swap out the file atomically so clients watching the file never catch
        # us mid-write.
        logger.info("Updating %r", self.dest)
        Path(self.dest).parent.mkdir(parents=True, exist_ok=True)
        with open(self.dest + ".tmp", "wb") as tmpfile:
            if self.owner and self.group:
                os.fchown(tmpfile.fileno(), self.owner, self.group)
            os.fchmod(tmpfile.fileno(), self.mode)

            tmpfile.write(data)
        os.rename(self.dest + ".tmp", self.dest)


def _parse_loader_type(data: bytes) -> LoaderType:
    try:
        json_data = json.loads(data.decode("UTF-8"))
    except (UnicodeDecodeError, json.decoder.JSONDecodeError):
        logger.debug("Data is not parseable as JSON, loading as PASSTHROUGH")
        return LoaderType.PASSTHROUGH

    try:
        loader_type = json_data["live_data_watcher_load_type"]
    except (KeyError, AttributeError, TypeError):
        logger.debug(
            "Expected dict (JSON object) but got %s. Loading as PASSTHROUGH", str(type(json_data))
        )
        return LoaderType.PASSTHROUGH

    return LoaderType(loader_type)


def _load_from_s3(data: bytes) -> bytes:
    # While many of the baseplate configurations use an ini format,
    # we've opted for json in these internal-to-znode-configs because
    # we want them to be fully controlled by the writer of the znode
    # and json is an easier format for znode authors to work with.
    loader_config = json.loads(data.decode("UTF-8"))
    try:
        region_name = loader_config["region_name"]
        s3_kwargs = {
            "Bucket": loader_config["bucket_name"],
            "Key": loader_config["file_key"],
            "SSECustomerKey": loader_config["sse_key"],
            "SSECustomerAlgorithm": "AES256",
        }
    except KeyError as e:
        # We require all of these keys to properly read from S3.
        logger.exception(
            "Failed to update live config: unable to fetch content from s3: source config has invalid or missing keys: %s.",
            e.args[0],
        )
        raise LoaderException from e

    if loader_config.get("anon") is True:
        # Client needs to be anonymous/unsigned or boto3 will try to read the local credentials
        # on the service pods. And - due to an AWS quirk - any request that comes in signed with
        # credentials will profile for permissions for the resource being requested EVEN if the
        # resource is public. In other words, this means that a given service cannot access
        # a public resource belonging to another cluster/AWS account unless the request credentials
        # are unsigned.
        s3_client = boto3.client(
            "s3",
            config=Config(signature_version=UNSIGNED),
            region_name=region_name,
        )
    else:
        s3_client = boto3.client(
            "s3",
            region_name=region_name,
        )

    try:
        s3_object = s3_client.get_object(**s3_kwargs)
        # Returns bytes.
        return s3_object["Body"].read()
    except ClientError as error:
        logger.exception(
            "Failed to update live config: failed to load data from S3. Received error code %s: %s",
            error.response["Error"]["Code"],
            error.response["Error"]["Message"],
        )

        raise LoaderException from error
    except EndpointConnectionError as error:
        logger.exception("Unable to retrieve object")
        raise LoaderException from error
    except ValueError as error:
        logger.exception(
            "Failed to update live config: params for loading from S3 are incorrect. Received error: %s",
            error,
        )

        raise LoaderException from error


def watch_zookeeper_nodes(zookeeper: KazooClient, nodes: Any) -> NoReturn:
    for node in nodes:
        watcher = NodeWatcher(node.dest, node.owner, node.group, node.mode)
        zookeeper.DataWatch(node.source, watcher.on_change)

    # all the interesting stuff is now happening in the Kazoo worker thread
    # and so we'll just spin and periodically heartbeat to prove we're alive.
    while True:
        time.sleep(HEARTBEAT_INTERVAL)

        # see the comment in baseplate.live_data.zookeeper for explanation of
        # how reconnects work with the background thread.
        if zookeeper.connected:
            for node in nodes:
                try:
                    logger.debug("Heartbeating %s", node.dest)

                    # this will make FileWatchers re-parse the file on the next
                    # read which is unfortunate but we do it anyway. it's
                    # important to monitor that the file is being updated as
                    # accurately as possible rather than using a separate file
                    # or mechanism as a proxy. for example, the actual output
                    # file could have bogus permissions that would go unnoticed
                    # if the heartbeat still worked independently.
                    os.utime(node.dest, None)
                except OSError as exc:
                    logger.warning("%s: could not heartbeat: %s", node.dest, exc)


def main() -> NoReturn:
    arg_parser = argparse.ArgumentParser(description=sys.modules[__name__].__doc__)
    arg_parser.add_argument(
        "config_file", type=argparse.FileType("r"), help="path to a configuration file"
    )
    arg_parser.add_argument(
        "--debug", default=False, action="store_true", help="enable debug logging"
    )
    args = arg_parser.parse_args()

    if args.debug:
        level = logging.DEBUG
    else:
        level = logging.INFO
    logging.basicConfig(level=level, format="%(message)s")

    # quiet kazoo's verbose logs a bit
    logging.getLogger("kazoo").setLevel(logging.WARNING)

    parser = configparser.RawConfigParser(interpolation=EnvironmentInterpolation())
    parser.read_file(args.config_file)
    watcher_config = dict(parser.items("live-data"))

    cfg = config.parse_config(
        watcher_config,
        {
            "nodes": config.DictOf(
                {
                    "source": config.String,
                    "dest": config.String,
                    "owner": config.Optional(config.UnixUser),
                    "group": config.Optional(config.UnixGroup),
                    "mode": config.Optional(config.Integer(base=8), default=0o400),  # type: ignore
                }
            )
        },
    )
    # pylint: disable=maybe-no-member
    nodes = cfg.nodes.values()

    secrets = secrets_store_from_config(watcher_config, timeout=30)
    zookeeper = zookeeper_client_from_config(secrets, watcher_config, read_only=True)
    zookeeper.start()
    try:
        watch_zookeeper_nodes(zookeeper, nodes)
    finally:
        zookeeper.stop()


if __name__ == "__main__":
    main()



import socket

from queue import Queue
from unittest import mock

import kombu
import pytest

from gevent.server import StreamServer
from prometheus_client import REGISTRY

from baseplate import Baseplate
from baseplate import RequestContext
from baseplate import ServerSpan
from baseplate.frameworks.queue_consumer.kombu import AMQP_ACTIVE_MESSAGES
from baseplate.frameworks.queue_consumer.kombu import AMQP_PROCESSED_TOTAL
from baseplate.frameworks.queue_consumer.kombu import AMQP_PROCESSING_TIME
from baseplate.frameworks.queue_consumer.kombu import AmqpConsumerPrometheusLabels
from baseplate.frameworks.queue_consumer.kombu import FatalMessageHandlerError
from baseplate.frameworks.queue_consumer.kombu import KombuConsumerWorker
from baseplate.frameworks.queue_consumer.kombu import KombuMessageHandler
from baseplate.frameworks.queue_consumer.kombu import KombuQueueConsumerFactory

from .... import does_not_raise


@pytest.fixture
def context():
    return mock.Mock(spec=RequestContext)


@pytest.fixture
def span():
    return mock.Mock(spec=ServerSpan)


@pytest.fixture
def baseplate(context, span):
    bp = mock.MagicMock(spec=Baseplate)
    bp._metrics_client = None
    bp.make_server_span().__enter__.return_value = span
    bp.make_context_object.return_value = context
    # Reset the mock calls since setting up the span actually triggers a "call"
    # to bp.make_server_span
    bp.reset_mock()
    return bp


@pytest.fixture
def name():
    return "test-handler"


class TestKombuMessageHandler:
    def setup(self):
        AMQP_PROCESSING_TIME.clear()
        AMQP_PROCESSED_TOTAL.clear()
        AMQP_ACTIVE_MESSAGES.clear()

    @pytest.fixture
    def message(self, connection):
        msg = mock.Mock(spec=kombu.Message)
        msg.delivery_info = {
            "routing_key": "routing-key",
            "consumer_tag": "consumer-tag",
            "delivery_tag": "delivery-tag",
            "exchange": "exchange",
        }
        msg.channel.connection.client = connection
        msg.decode.return_value = {"foo": "bar"}
        return msg

    def test_handle(self, context, span, baseplate, name, message):
        handler_fn = mock.Mock()
        handler = KombuMessageHandler(baseplate, name, handler_fn)
        prom_labels = AmqpConsumerPrometheusLabels(
            amqp_address="hostname:port",
            amqp_virtual_host="/",
            amqp_exchange_name="exchange",
            amqp_routing_key="routing-key",
        )
        handler.handle(message)

        baseplate.make_context_object.assert_called_once()
        baseplate.make_server_span.assert_called_once_with(context, name)
        baseplate.make_server_span().__enter__.assert_called_once()
        span.set_tag.assert_has_calls(
            [
                mock.call("kind", "consumer"),
                mock.call("amqp.routing_key", "routing-key"),
                mock.call("amqp.consumer_tag", "consumer-tag"),
                mock.call("amqp.delivery_tag", "delivery-tag"),
                mock.call("amqp.exchange", "exchange"),
            ],
            any_order=True,
        )
        handler_fn.assert_called_once_with(context, message.decode(), message)
        message.ack.assert_called_once()
        assert (
            REGISTRY.get_sample_value(
                f"{AMQP_PROCESSING_TIME._name}_bucket",
                {**prom_labels._asdict(), **{"amqp_success": "true", "le": "+Inf"}},
            )
            == 1
        )
        assert (
            REGISTRY.get_sample_value(
                f"{AMQP_PROCESSED_TOTAL._name}_total",
                {**prom_labels._asdict(), **{"amqp_success": "true"}},
            )
            == 1
        )
        assert (
            REGISTRY.get_sample_value(f"{AMQP_ACTIVE_MESSAGES._name}", prom_labels._asdict()) == 0
        )

    @pytest.mark.parametrize(
        "err,expectation",
        [
            (ValueError(), does_not_raise()),
            (FatalMessageHandlerError(), pytest.raises(FatalMessageHandlerError)),
        ],
    )
    def test_errors(self, err, expectation, baseplate, name, message):
        def handler_fn(ctx, body, msg):
            raise err

        prom_labels = AmqpConsumerPrometheusLabels(
            amqp_address="hostname:port",
            amqp_virtual_host="/",
            amqp_exchange_name="exchange",
            amqp_routing_key="routing-key",
        )
        mock_manager = mock.Mock()
        with mock.patch.object(
            AMQP_ACTIVE_MESSAGES.labels(**prom_labels._asdict()),
            "inc",
            wraps=AMQP_ACTIVE_MESSAGES.labels(**prom_labels._asdict()).inc,
        ) as active_inc_spy_method:
            mock_manager.attach_mock(active_inc_spy_method, "inc")
            with mock.patch.object(
                AMQP_ACTIVE_MESSAGES.labels(**prom_labels._asdict()),
                "dec",
                wraps=AMQP_ACTIVE_MESSAGES.labels(**prom_labels._asdict()).dec,
            ) as active_dec_spy_method:
                mock_manager.attach_mock(active_dec_spy_method, "dec")
                handler = KombuMessageHandler(baseplate, name, handler_fn)
                with expectation:
                    handler.handle(message)
                message.requeue.assert_called_once()

                assert (
                    REGISTRY.get_sample_value(
                        f"{AMQP_PROCESSING_TIME._name}_bucket",
                        {**prom_labels._asdict(), **{"amqp_success": "false", "le": "+Inf"}},
                    )
                    == 1
                )
                assert (
                    REGISTRY.get_sample_value(
                        f"{AMQP_PROCESSED_TOTAL._name}_total",
                        {**prom_labels._asdict(), **{"amqp_success": "false"}},
                    )
                    == 1
                )
                assert (
                    REGISTRY.get_sample_value(
                        f"{AMQP_ACTIVE_MESSAGES._name}", prom_labels._asdict()
                    )
                    == 0
                )
                # we need to assert that not only the end result is 0, but that we increased and then decreased to that value
                assert mock_manager.mock_calls == [mock.call.inc(), mock.call.dec()]

    @pytest.mark.parametrize(
        "err,expectation",
        [
            (ValueError(), does_not_raise()),
            (FatalMessageHandlerError(), pytest.raises(FatalMessageHandlerError)),
        ],
    )
    def test_errors_with_error_handler_fn(
        self, err, expectation, context, baseplate, name, message
    ):
        def handler_fn(ctx, body, msg):
            raise err

        error_handler_fn = mock.Mock()

        prom_labels = AmqpConsumerPrometheusLabels(
            amqp_address="hostname:port",
            amqp_virtual_host="/",
            amqp_exchange_name="exchange",
            amqp_routing_key="routing-key",
        )
        mock_manager = mock.Mock()
        with mock.patch.object(
            AMQP_ACTIVE_MESSAGES.labels(**prom_labels._asdict()),
            "inc",
            wraps=AMQP_ACTIVE_MESSAGES.labels(**prom_labels._asdict()).inc,
        ) as active_inc_spy_method:
            mock_manager.attach_mock(active_inc_spy_method, "inc")
            with mock.patch.object(
                AMQP_ACTIVE_MESSAGES.labels(**prom_labels._asdict()),
                "dec",
                wraps=AMQP_ACTIVE_MESSAGES.labels(**prom_labels._asdict()).dec,
            ) as active_dec_spy_method:
                mock_manager.attach_mock(active_dec_spy_method, "dec")

                handler = KombuMessageHandler(baseplate, name, handler_fn, error_handler_fn)
                with expectation:
                    handler.handle(message)
                error_handler_fn.assert_called_once_with(context, message.decode(), message, err)
                message.ack.assert_not_called()
                message.requeue.assert_not_called()

                assert (
                    REGISTRY.get_sample_value(
                        f"{AMQP_PROCESSING_TIME._name}_bucket",
                        {**prom_labels._asdict(), **{"amqp_success": "false", "le": "+Inf"}},
                    )
                    == 1
                )
                assert (
                    REGISTRY.get_sample_value(
                        f"{AMQP_PROCESSED_TOTAL._name}_total",
                        {**prom_labels._asdict(), **{"amqp_success": "false"}},
                    )
                    == 1
                )
                assert (
                    REGISTRY.get_sample_value(
                        f"{AMQP_ACTIVE_MESSAGES._name}", prom_labels._asdict()
                    )
                    == 0
                )
                # we need to assert that not only the end result is 0, but that we increased and then decreased to that value
                assert mock_manager.mock_calls == [mock.call.inc(), mock.call.dec()]


@pytest.fixture
def connection():
    return mock.Mock(spec=kombu.Connection, virtual_host="/", host="hostname:port")


@pytest.fixture
def exchange():
    return mock.Mock(spec=kombu.Exchange)


@pytest.fixture
def routing_keys():
    return ["rk-1", "rk-2"]


class TestQueueConsumerFactory:
    @pytest.fixture
    def make_queue_consumer_factory(self, baseplate, exchange, connection, name, routing_keys):
        def _make_queue_consumer_factory(health_check_fn=None):
            worker_kwargs = {"prefetch_limit": 1}
            return KombuQueueConsumerFactory.new(
                baseplate=baseplate,
                exchange=exchange,
                connection=connection,
                queue_name=name,
                routing_keys=routing_keys,
                handler_fn=lambda ctx, body, msg: True,
                error_handler_fn=lambda ctx, body, msg: True,
                health_check_fn=health_check_fn,
                worker_kwargs=worker_kwargs,
            )

        return _make_queue_consumer_factory

    def test_new(self, baseplate, exchange, connection, name, routing_keys):
        handler_fn = mock.Mock()
        error_handler_fn = mock.Mock()
        health_check_fn = mock.Mock()
        worker_kwargs = {"prefetch_limit": 1}
        factory = KombuQueueConsumerFactory.new(
            baseplate=baseplate,
            exchange=exchange,
            connection=connection,
            queue_name=name,
            routing_keys=routing_keys,
            handler_fn=handler_fn,
            error_handler_fn=error_handler_fn,
            health_check_fn=health_check_fn,
            worker_kwargs=worker_kwargs,
        )
        assert factory.baseplate == baseplate
        assert factory.connection == connection
        assert factory.name == name
        assert factory.handler_fn == handler_fn
        assert factory.error_handler_fn == error_handler_fn
        assert factory.health_check_fn == health_check_fn
        for routing_key, queue in zip(routing_keys, factory.queues):
            assert queue.routing_key == routing_key
            assert queue.name == name
            assert queue.exchange == exchange
        assert worker_kwargs == worker_kwargs

    def test_build_pump_worker(self, make_queue_consumer_factory):
        factory = make_queue_consumer_factory()
        work_queue = Queue(maxsize=10)
        pump = factory.build_pump_worker(work_queue)
        assert isinstance(pump, KombuConsumerWorker)
        assert pump.connection == factory.connection
        assert pump.queues == factory.queues
        assert pump.work_queue == work_queue
        assert pump.kwargs == {"prefetch_limit": 1}

    def test_build_message_handler(self, make_queue_consumer_factory):
        factory = make_queue_consumer_factory()
        handler = factory.build_message_handler()
        assert isinstance(handler, KombuMessageHandler)
        assert handler.baseplate == factory.baseplate
        assert handler.name == factory.name
        assert handler.handler_fn == factory.handler_fn
        assert handler.error_handler_fn == factory.error_handler_fn

    @pytest.mark.parametrize("health_check_fn", [None, lambda req: True])
    def test_build_health_checker(self, health_check_fn, make_queue_consumer_factory):
        factory = make_queue_consumer_factory(health_check_fn=health_check_fn)
        listener = mock.Mock(spec=socket.socket)
        health_checker = factory.build_health_checker(listener)
        assert isinstance(health_checker, StreamServer)


@pytest.fixture
def queues(name, exchange, routing_keys):
    return [kombu.Queue(name=name, exchange=exchange, routing_key=key) for key in routing_keys]


class TestKombuConsumerWorker:
    @pytest.fixture
    def consumer_worker(self, connection, queues):
        return KombuConsumerWorker(connection, queues, Queue(maxsize=10))

    def test_initial_state(self, consumer_worker):
        assert getattr(consumer_worker, "should_stop", None) is not True

    @mock.patch("baseplate.frameworks.queue_consumer.kombu.ConsumerMixin.run")
    def test_run(self, run, consumer_worker):
        consumer_worker.run()
        run.assert_called_once()

    def test_stop(self, consumer_worker):
        consumer_worker.stop()
        assert consumer_worker.should_stop is True



"""Client library for sending events to the data processing system.

This is for use with the event collector system. Events generally track
something that happens in production that we want to instrument for planning
and analytical purposes.

Events are serialized and put onto a message queue on the same server. These
serialized events are then consumed and published to the remote event collector
by a separate daemon.

"""
import logging

from typing import Any
from typing import Callable
from typing import Generic
from typing import TypeVar

from thrift import TSerialization
from thrift.protocol.TJSONProtocol import TJSONProtocolFactory

from baseplate import Span
from baseplate.clients import ContextFactory
from baseplate.lib import config
from baseplate.lib.message_queue import MessageQueue
from baseplate.lib.message_queue import TimedOutError


MAX_EVENT_SIZE = 102400
MAX_QUEUE_SIZE = 10000


class EventError(Exception):
    """Base class for event related exceptions."""


class EventTooLargeError(EventError):
    """Raised when a serialized event is too large to send."""

    def __init__(self, size: int):
        super().__init__(f"Event is too large to send ({size:d} bytes)")


class EventQueueFullError(EventError):
    """Raised when the queue of events is full.

    This usually indicates that the event publisher is having trouble talking
    to the event collector.

    """

    def __init__(self) -> None:
        super().__init__("The event queue is full.")


_V2_PROTOCOL_FACTORY = TJSONProtocolFactory()


def serialize_v2_event(event: Any) -> bytes:
    """Serialize a Thrift struct to bytes for the V2 event protocol.

    :param event: A Thrift struct from the event schemas.

    """
    return TSerialization.serialize(event, _V2_PROTOCOL_FACTORY)


class EventLogger:
    def log(self, **kwargs: Any) -> None:
        raise NotImplementedError


class DebugLogger(EventLogger):
    def __init__(self) -> None:
        self.logger = logging.getLogger(__name__)

    def log(self, **kwargs: Any) -> None:
        self.logger.debug("Would send event: %s", kwargs)


T = TypeVar("T")


class EventQueue(ContextFactory, config.Parser, Generic[T]):
    """A queue to transfer events to the publisher.

    :param name: The name of the event queue to send to. This specifies
        which publisher should send the events which can be useful for routing
        to different event pipelines (prod/test/v2 etc.).
    :param event_serializer: A callable that takes an event object
        and returns serialized bytes ready to send on the wire. See below for
        options.

    """

    def __init__(self, name: str, event_serializer: Callable[[T], bytes]):
        self.queue = MessageQueue(
            "/events-" + name, max_messages=MAX_QUEUE_SIZE, max_message_size=MAX_EVENT_SIZE
        )
        self.serialize_event = event_serializer

    def put(self, event: T) -> None:
        """Add an event to the queue.

        The queue is local to the server this code is run on. The event
        publisher on the server will take these events and send them to the
        collector.

        :param event: The event to send. The type of event object passed in
            depends on the selected ``event_serializer``.
        :raises: :py:exc:`EventTooLargeError` The serialized event is too large.
        :raises: :py:exc:`EventQueueFullError` The queue is full. Events are
            not being published fast enough.

        """
        serialized = self.serialize_event(event)
        if len(serialized) > MAX_EVENT_SIZE:
            raise EventTooLargeError(len(serialized))

        try:
            self.queue.put(serialized, timeout=0)
        except TimedOutError:
            raise EventQueueFullError

    def make_object_for_context(self, name: str, span: Span) -> "EventQueue[T]":
        return self

    def parse(self, key_path: str, raw_config: config.RawConfig) -> "EventQueue[T]":
        return self



import datetime

from unittest import mock

import pytest

from baseplate.lib import crypto
from baseplate.lib.secrets import VersionedSecret


TEST_SECRET = VersionedSecret(previous=b"one", current=b"two", next=b"three")
MESSAGE = "test message"
VALID_TIL_1030 = b"AQAABgQAAOMD6M5zvQU0-GK_uKvPdKH7NOeRAq5Jdlkjwq67BzLt"


@mock.patch("time.time")
def test_make_signature(mock_time):
    mock_time.return_value = 1000

    signature = crypto.make_signature(TEST_SECRET, MESSAGE, max_age=datetime.timedelta(seconds=30))

    assert signature == VALID_TIL_1030


@pytest.mark.parametrize(
    "signature",
    [
        b"totally bogus",
        b"Ym9ndXM=",  # base64, but "bogus" content
        b"AgAA0gQAAGFzZGZhc2Rm",  # v2 header
        b"AQAABgQAAOMD6M5zvQU0",  # wrong length
    ],
)
def test_bogus_signature(signature):
    with pytest.raises(crypto.UnreadableSignatureError):
        crypto.validate_signature(TEST_SECRET, MESSAGE, signature)


@mock.patch("time.time")
def test_expired(mock_time):
    mock_time.return_value = 2000

    with pytest.raises(crypto.ExpiredSignatureError) as exc:
        crypto.validate_signature(TEST_SECRET, MESSAGE, VALID_TIL_1030)

    assert exc.value.expiration == 1030


@mock.patch("time.time")
@pytest.mark.parametrize(
    "rotated_secret",
    [
        VersionedSecret(previous=TEST_SECRET.current, current=b"new", next=b"new"),
        VersionedSecret(previous=b"old", current=TEST_SECRET.current, next=b"new"),
        VersionedSecret(previous=b"old", current=b"old", next=TEST_SECRET.current),
    ],
)
def test_secret_rotation(mock_time, rotated_secret):
    mock_time.return_value = 1000

    result = crypto.validate_signature(rotated_secret, MESSAGE, VALID_TIL_1030)

    assert result.version == 1
    assert result.expiration == 1030


@mock.patch("time.time")
def test_bad_signature(mock_time):
    mock_time.return_value = 1000

    with pytest.raises(crypto.IncorrectSignatureError):
        crypto.validate_signature(TEST_SECRET, "SNEAKY DIFFERENT MESSAGE", VALID_TIL_1030)



"""Automatic reloader for development environments.

This will watch all active source files and the application configuration and
restart the app if anything changes. This should not be used in production
settings.

"""
import logging
import os
import re
import sys
import threading
import time

from typing import Dict
from typing import Iterator
from typing import NoReturn
from typing import Sequence


logger = logging.getLogger(__name__)


def _get_loaded_modules() -> Iterator[str]:
    """Yield filenames for all loaded Python modules."""
    for module in sys.modules.values():
        filename = getattr(module, "__file__", None)
        if filename:
            uncompiled = re.sub("py[co]$", "py", filename)
            yield uncompiled


def _get_watched_files(extra_files: Sequence[str]) -> Iterator[str]:
    """Yield filenames for all files to be watched for modification."""
    yield from _get_loaded_modules()
    yield from extra_files


def _reload_when_files_change(extra_files: Sequence[str]) -> NoReturn:
    """Scan all watched files periodically and re-exec if anything changed."""
    initial_mtimes: Dict[str, float] = {}
    while True:
        for filename in _get_watched_files(extra_files):
            try:
                current_mtime = os.path.getmtime(filename)
            except os.error:
                continue

            initial_mtimes.setdefault(filename, current_mtime)
            if initial_mtimes[filename] < current_mtime:
                logger.debug("Reloading, %s changed", filename)
                os.execl(sys.executable, sys.executable, *sys.argv)

        time.sleep(0.25)


def start_reload_watcher(extra_files: Sequence[str]) -> None:
    """Start a task that will restart the server if any source files change."""
    thread = threading.Thread(target=_reload_when_files_change, args=(extra_files,))
    thread.name = "baseplate reloader"
    thread.daemon = True
    thread.start()



import logging
import os
import random

from contextlib import contextmanager
from types import TracebackType
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterator
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Tuple
from typing import Type

import gevent.monkey

from pkg_resources import DistributionNotFound
from pkg_resources import get_distribution

from baseplate.lib import config
from baseplate.lib import get_calling_module_name
from baseplate.lib import metrics
from baseplate.lib import UnknownCallerError


try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    __version__ = "unknown"


logger = logging.getLogger(__name__)


class BaseplateObserver:
    """Interface for an observer that watches Baseplate."""

    def on_server_span_created(self, context: "RequestContext", server_span: "ServerSpan") -> None:
        """Do something when a server span is created.

        :py:class:`Baseplate` calls this when a new request begins.

        :param context: The :py:class:`~baseplate.RequestContext` for this
            request.
        :param server_span: The span representing this request.

        """
        raise NotImplementedError


_ExcInfo = Tuple[Optional[Type[BaseException]], Optional[BaseException], Optional[TracebackType]]


class SpanObserver:
    """Interface for an observer that watches a span."""

    def on_start(self) -> None:
        """Do something when the observed span is started."""

    def on_set_tag(self, key: str, value: Any) -> None:
        """Do something when a tag is set on the observed span."""

    def on_incr_tag(self, key: str, delta: float) -> None:
        """Do something when a tag value is incremented on the observed span."""

    def on_log(self, name: str, payload: Any) -> None:
        """Do something when a log entry is added to the span."""

    def on_finish(self, exc_info: Optional[_ExcInfo]) -> None:
        """Do something when the observed span is finished.

        :param exc_info: If the span ended because of an exception, the
            exception info. Otherwise, :py:data:`None`.

        """

    def on_child_span_created(self, span: "Span") -> None:
        """Do something when a child span is created.

        :py:class:`SpanObserver` objects call this when a new child span is
        created.

        :param span: The new child span.

        """


class ServerSpanObserver(SpanObserver):
    """Interface for an observer that watches the server span."""


class TraceInfo(NamedTuple):
    """Trace context for a span.

    If this request was made at the behest of an upstream service, the upstream
    service should have passed along trace information. This class is used for
    collecting the trace context and passing it along to the server span.

    """

    #: The ID of the whole trace. This will be the same for all downstream requests.
    trace_id: str

    #: The ID of the parent span, or None if this is the root span.
    parent_id: Optional[str]

    #: The ID of the current span. Should be unique within a trace.
    span_id: str

    #: True if this trace was selected for sampling. Will be propagated to child spans.
    sampled: Optional[bool]

    #: A bit field of extra flags about this trace.
    flags: Optional[int]

    @classmethod
    def new(cls) -> "TraceInfo":
        """Generate IDs for a new initial server span.

        This span has no parent and has a random ID. It cannot be correlated
        with any upstream requests.

        """
        trace_id = str(random.getrandbits(64))
        return cls(trace_id=trace_id, parent_id=None, span_id=trace_id, sampled=None, flags=None)

    @classmethod
    def from_upstream(
        cls,
        trace_id: Optional[str],
        parent_id: Optional[str],
        span_id: Optional[str],
        sampled: Optional[bool],
        flags: Optional[int],
    ) -> "TraceInfo":
        """Build a TraceInfo from individual headers.

        :param trace_id: The ID of the trace.
        :param parent_id: The ID of the parent span.
        :param span_id: The ID of this span within the tree.
        :param sampled: Boolean flag to determine request sampling.
        :param flags: Bit flags for communicating feature flags downstream

        :raises: :py:exc:`ValueError` if any of the values are inappropriate.

        """
        if trace_id is None:
            raise ValueError("invalid trace_id")

        if span_id is None:
            raise ValueError("invalid span_id")

        if parent_id is None:
            raise ValueError("invalid parent_id")

        if sampled is not None and not isinstance(sampled, bool):
            raise ValueError("invalid sampled value")

        if flags is not None:
            if not 0 <= flags < 2 ** 64:
                raise ValueError("invalid flags value")

        return cls(trace_id, parent_id, span_id, sampled, flags)


class RequestContext:
    """The request context object.

    The context object is passed into each request handler by the framework
    you're using. In some cases (e.g. Pyramid) the request object will also
    inherit from another base class and carry extra framework-specific
    information.

    Clients and configuration added to the context via
    :py:meth:`~baseplate.Baseplate.configure_context` or
    :py:meth:`~baseplate.Baseplate.add_to_context` will be available as an
    attribute on this object.  To take advantage of Baseplate's automatic
    monitoring, any interactions with external services should be done through
    these clients.

    """

    def __init__(
        self,
        context_config: Dict[str, Any],
        prefix: Optional[str] = None,
        span: Optional["Span"] = None,
        wrapped: Optional["RequestContext"] = None,
    ):
        self.__context_config = context_config
        self.__prefix = prefix
        self.__wrapped = wrapped

        # the context and span reference eachother (unfortunately) so we can't
        # construct 'em both with references from the start. however, we can
        # guarantee that during the valid life of a span, there will be a
        # reference. so we fake it here and say "trust us".
        #
        # this would be much cleaner with a different API but this is where we are.
        self.span: "Span" = span  # type: ignore

    def __getattr__(self, name: str) -> Any:
        try:
            config_item = self.__context_config[name]
        except KeyError:
            try:
                return getattr(self.__wrapped, name)
            except AttributeError:
                raise AttributeError(
                    f"{repr(self.__class__.__name__)} object has no attribute {repr(name)}"
                ) from None

        if self.__prefix:
            full_name = f"{self.__prefix}.{name}"
        else:
            full_name = name

        if isinstance(config_item, dict):
            obj = RequestContext(context_config=config_item, prefix=full_name, span=self.span)
        elif hasattr(config_item, "make_object_for_context"):
            obj = config_item.make_object_for_context(full_name, self.span)
        else:
            obj = config_item

        setattr(self, name, obj)
        return obj

    # this is just here for type checking
    # pylint: disable=useless-super-delegation
    def __setattr__(self, name: str, value: Any) -> None:
        super().__setattr__(name, value)

    def clone(self) -> "RequestContext":
        return RequestContext(
            context_config=self.__context_config,
            prefix=self.__prefix,
            span=self.span,
            wrapped=self,
        )


class ReusedContextObjectError(Exception):
    def __init__(self) -> None:
        super().__init__(
            "Context objects cannot be re-used. See https://baseplate.readthedocs.io/en/latest/guide/faq.html#what-do-i-do-about-context-objects-cannot-be-re-used"
        )


class Baseplate:
    """The core of the Baseplate framework.

    This class coordinates monitoring and tracing of service calls made to
    and from this service. See :py:mod:`baseplate.frameworks` for how to
    integrate it with the application framework you are using.

    """

    def __init__(self, app_config: Optional[config.RawConfig] = None) -> None:
        """Initialize the core observability framework.

        :param app_config: The raw configuration dictionary for your
            application as supplied by :program:`baseplate-serve` or
            :program:`baseplate-script`. In addition to allowing
            framework-level configuration (described next), if this is supplied
            you do not need to pass the configuration again when calling
            :py:meth:`configure_observers` or :py:meth:`configure_context`.

        Baseplate services can identify themselves to downstream services in
        requests. The name a service identifies as defaults to the Python
        module the :py:class:`~baseplate.Baseplate` object is instantiated in.
        To override the default, make sure you are passing in an `app_config`
        and configure the name in your INI file:

        .. code-block:: ini

            [app:main]
            baseplate.service_name = foo_service

            ...

        """
        self.observers: List[BaseplateObserver] = []
        self._metrics_client: Optional[metrics.Client] = None
        self._context_config: Dict[str, Any] = {}
        self._app_config = app_config or {}

        self.service_name = self._app_config.get("baseplate.service_name")
        if not self.service_name:
            try:
                self.service_name = get_calling_module_name()
            except UnknownCallerError:
                # this happens e.g. when instantiating Baseplate() in a shell
                pass

    def register(self, observer: BaseplateObserver) -> None:
        """Register an observer.

        :param observer: An observer.

        """
        self.observers.append(observer)

    # pylint: disable=cyclic-import
    def configure_observers(self) -> None:
        """Configure diagnostics observers based on application configuration.

        This installs all the currently supported observers that have settings
        in the configuration file.

        See :py:mod:`baseplate.observers` for the configuration settings
        available for each observer.

        """
        skipped = []

        from baseplate.observers.logging import LoggingBaseplateObserver

        self.register(LoggingBaseplateObserver())

        if gevent.monkey.is_module_patched("socket"):
            from baseplate.observers.timeout import TimeoutBaseplateObserver

            timeout_observer = TimeoutBaseplateObserver.from_config(self._app_config)
            self.register(timeout_observer)
        else:
            skipped.append("timeout")

        if "metrics.tagging" in self._app_config:
            if "metrics.namespace" in self._app_config:
                raise ValueError("metrics.namespace not allowed with metrics.tagging")
            from baseplate.lib.metrics import metrics_client_from_config
            from baseplate.observers.metrics_tagged import TaggedMetricsBaseplateObserver

            self._metrics_client = metrics_client_from_config(self._app_config)
            self.register(
                TaggedMetricsBaseplateObserver.from_config_and_client(
                    self._app_config, self._metrics_client
                )
            )

        elif "metrics.namespace" in self._app_config:
            from baseplate.lib.metrics import metrics_client_from_config
            from baseplate.observers.metrics import MetricsBaseplateObserver

            self._metrics_client = metrics_client_from_config(self._app_config)
            self.register(
                MetricsBaseplateObserver.from_config_and_client(
                    self._app_config, self._metrics_client
                )
            )

        else:
            skipped.append("metrics")

        if "tracing.service_name" in self._app_config:
            from baseplate.observers.tracing import tracing_client_from_config
            from baseplate.observers.tracing import TraceBaseplateObserver

            tracing_client = tracing_client_from_config(self._app_config)
            self.register(TraceBaseplateObserver(tracing_client))
        else:
            skipped.append("tracing")

        if "sentry.dsn" in self._app_config or "SENTRY_DSN" in os.environ:
            from baseplate.observers.sentry import init_sentry_client_from_config
            from baseplate.observers.sentry import SentryBaseplateObserver
            from baseplate.observers.sentry import _SentryUnhandledErrorReporter

            init_sentry_client_from_config(self._app_config)
            _SentryUnhandledErrorReporter.install()
            self.register(SentryBaseplateObserver())
        else:
            skipped.append("sentry")

        if skipped:
            logger.debug(
                "The following observers are unconfigured and won't run: %s", ", ".join(skipped)
            )

    def configure_context(self, context_spec: Dict[str, Any]) -> None:
        """Add a number of objects to each request's context object.

        Configure and attach multiple clients to the
        :py:class:`~baseplate.RequestContext` in one place. This takes a full
        configuration spec like :py:func:`baseplate.lib.config.parse_config`
        and will attach the specified structure onto the context object each
        request.

        For example, a configuration like::

            baseplate = Baseplate(app_config)
            baseplate.configure_context({
                "cfg": {
                    "doggo_is_good": config.Boolean,
                },
                "cache": MemcachedClient(),
                "cassandra": {
                    "foo": CassandraClient("foo_keyspace"),
                    "bar": CassandraClient("bar_keyspace"),
                },
            })

        would build a context object that could be used like::

            assert context.cfg.doggo_is_good == True
            context.cache.get("example")
            context.cassandra.foo.execute()

        :param app_config: The raw stringy configuration dictionary.
        :param context_spec: A specification of what the configuration should
            look like.

        """
        cfg = config.parse_config(self._app_config, context_spec)
        self._context_config.update(cfg)

    def add_to_context(self, name: str, attribute_config: Any) -> None:
        """Add an attribute or a structure of attributes to each request's context object.

        The given attribute config object can be one of the following:

        * An arbitrary object to be added to the :py:class:`~baseplate.RequestContext`.

        * A factory with a method named ``make_object_for_context``.  On each
          request, the factory will be asked to create an appropriate object to
          attach to the :py:class:`~baseplate.RequestContext`.

        * A dict containing arbitrary objects, factories, or other dicts.  In
          this case, a nested object will be added to the context. Each item of
          the dict will be processed using the same rules to become an attribute
          of the nested object.

        :param name: The attribute on the context object to attach the
            created object to. This may also be used for metric/tracing
            purposes so it should be descriptive.
        :param attribute_config: A configuration object.

        """
        self._context_config[name] = attribute_config

    def make_context_object(self) -> RequestContext:
        """Make a context object for the request."""
        return RequestContext(self._context_config)

    def make_server_span(
        self, context: RequestContext, name: str, trace_info: Optional[TraceInfo] = None
    ) -> "ServerSpan":
        """Return a server span representing the request we are handling.

        In a server, a server span represents the time spent on a single
        incoming request. Any calls made to downstream services will be new
        child spans of the server span, and the server span will in turn be the
        child span of whatever upstream request it is part of, if any.

        :param context: The :py:class:`~baseplate.RequestContext` for this
            request.
        :param name: A name to identify the type of this request, e.g.  a route
            or RPC method name.
        :param trace_info: The trace context of this request as passed in from
            upstream. If :py:data:`None`, a new trace context will be generated.

        """
        assert isinstance(context, RequestContext)

        if context.span is not None:
            raise ReusedContextObjectError

        if trace_info is None:
            trace_info = TraceInfo.new()

        server_span = ServerSpan(
            trace_info.trace_id,
            trace_info.parent_id,
            trace_info.span_id,
            trace_info.sampled,
            trace_info.flags,
            name,
            context,
            baseplate=self,
        )
        context.span = server_span

        for observer in self.observers:
            observer.on_server_span_created(context, server_span)
        return server_span

    @contextmanager
    def server_context(self, name: str) -> Iterator[RequestContext]:
        """Create a server span and return a context manager for its lifecycle.

        This is a convenience wrapper around a common pattern seen outside of
        servers handling requests. For example, simple cron jobs or one-off
        scripts might want to create a temporary span and access the context
        object. Instead of calling
        :py:meth:`~baseplate.Baseplate.make_context_object` followed by
        :py:meth:`~baseplate.Baseplate.make_server_span` manually, this method
        bundles it all up for you::

            with baseplate.server_context("foo") as context:
                context.redis.ping()

        .. note::

            This should not be used within an existing span context (such as
            during request processing) as it creates a new span unrelated to
            any other ones.

        """
        context = self.make_context_object()
        with self.make_server_span(context, name):
            yield context

    def get_runtime_metric_reporters(self) -> Dict[str, Callable[[Any], None]]:
        specs: List[Tuple[Optional[str], Dict[str, Any]]] = [(None, self._context_config)]
        result = {}
        while specs:
            prefix, spec = specs.pop(0)
            for name, value in spec.items():
                if prefix:
                    full_name = f"{prefix}.{name}"
                else:
                    full_name = name

                if isinstance(value, dict):
                    specs.append((full_name, value))
                elif hasattr(value, "report_runtime_metrics"):
                    result[full_name] = value.report_runtime_metrics
        return result


class Span:
    """A span represents a single RPC within a system."""

    def __init__(
        self,
        trace_id: str,
        parent_id: Optional[str],
        span_id: str,
        sampled: Optional[bool],
        flags: Optional[int],
        name: str,
        context: RequestContext,
        baseplate: Optional[Baseplate] = None,
    ):
        self.trace_id = trace_id
        self.parent_id = parent_id
        self.id = span_id
        self.sampled = sampled
        self.flags = flags
        self.name = name
        self.context = context
        self.baseplate = baseplate
        self.component_name: Optional[str] = None
        self.observers: List[SpanObserver] = []

    def register(self, observer: SpanObserver) -> None:
        """Register an observer to receive events from this span."""
        self.observers.append(observer)

    def start(self) -> None:
        """Record the start of the span.

        This notifies any observers that the span has started, which indicates
        that timers etc. should start ticking.

        Spans also support the `context manager protocol`_, for use with
        Python's ``with`` statement. When the context is entered, the span
        calls :py:meth:`start` and when the context is exited it automatically
        calls :py:meth:`finish`.

        .. _context manager protocol:
            https://docs.python.org/3/reference/datamodel.html#context-managers

        """
        for observer in self.observers:
            observer.on_start()

    def set_tag(self, key: str, value: Any) -> None:
        """Set a tag on the span.

        Tags are arbitrary key/value pairs that add context and meaning to the
        span, such as a hostname or query string. Observers may interpret or
        ignore tags as they desire.

        :param key: The name of the tag.
        :param value: The value of the tag.

        """
        for observer in self.observers:
            observer.on_set_tag(key, value)

    def incr_tag(self, key: str, delta: float = 1) -> None:
        """Increment a tag value on the span.

        This is useful to count instances of an event in your application. In
        addition to showing up as a tag on the span, the value may also be
        aggregated separately as an independent counter.

        :param key: The name of the tag.
        :param value: The amount to increment the value. Defaults to 1.

        """
        for observer in self.observers:
            observer.on_incr_tag(key, delta)

    def log(self, name: str, payload: Optional[Any] = None) -> None:
        """Add a log entry to the span.

        Log entries are timestamped events recording notable moments in the
        lifetime of a span.

        :param name: The name of the log entry. This should be a stable
            identifier that can apply to multiple span instances.
        :param payload: Optional log entry payload. This can be arbitrary data.

        """
        for observer in self.observers:
            observer.on_log(name, payload)

    def finish(self, exc_info: Optional[_ExcInfo] = None) -> None:
        """Record the end of the span.

        :param exc_info: If the span ended because of an exception, this is
            the exception information. The default is :py:data:`None` which
            indicates normal exit.

        """
        for observer in self.observers:
            observer.on_finish(exc_info)

        # clean up reference cycles
        self.context = None  # type: ignore
        self.observers.clear()

    def __enter__(self) -> "Span":
        self.start()
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        if exc_type is not None:
            self.finish(exc_info=(exc_type, value, traceback))
        else:
            self.finish()

    def make_child(
        self, name: str, local: bool = False, component_name: Optional[str] = None
    ) -> "Span":
        """Return a child Span whose parent is this Span."""
        raise NotImplementedError

    def with_tags(self, tags: Dict[str, Any]) -> "Span":
        """Declare a set of tags to be added to a span before starting it in the context manager.

        Can be used as follow:
        tags = {...}
        with self.span.make_child("...").with_tags(tags) as span:
            ...

        :param tags: Dict of tags to be set on the span at creation time.
        `"""
        for k, v in tags.items():
            self.set_tag(k, v)
        return self


class ParentSpanAlreadyFinishedError(Exception):
    def __init__(self) -> None:
        super().__init__(
            "Cannot make child span of parent that already finished. See https://baseplate.readthedocs.io/en/latest/guide/faq.html#what-do-i-do-about-cannot-make-child-span-of-parent-that-already-finished"
        )


class LocalSpan(Span):
    def make_child(
        self, name: str, local: bool = False, component_name: Optional[str] = None
    ) -> "Span":
        """Return a child Span whose parent is this Span.

        The child span can either be a local span representing an in-request
        operation or a span representing an outbound service call.

        In a server, a local span represents the time spent within a
        local component performing an operation or set of operations.
        The local component is some grouping of business logic,
        which is then split up into operations which could each be wrapped
        in local spans.

        :param name: Name to identify the operation this span
            is recording.
        :param local: Make this span a LocalSpan if True, otherwise
            make this span a base Span.
        :param component_name: Name to identify local component
            this span is recording in if it is a local span.
        """
        if not self.context:
            raise ParentSpanAlreadyFinishedError

        span_id = str(random.getrandbits(64))
        context_copy = self.context.clone()
        span: Span
        if local:
            span = LocalSpan(
                self.trace_id,
                self.id,
                span_id,
                self.sampled,
                self.flags,
                name,
                context_copy,
                self.baseplate,
            )
            span.component_name = component_name
        else:
            span = Span(
                self.trace_id,
                self.id,
                span_id,
                self.sampled,
                self.flags,
                name,
                context_copy,
                self.baseplate,
            )
        context_copy.span = span

        for observer in self.observers:
            observer.on_child_span_created(span)
        return span


class ServerSpan(LocalSpan):
    """A server span represents a request this server is handling.

    The server span is available on the :py:class:`~baseplate.RequestContext`
    during requests as the ``span`` attribute.

    """


__all__ = ["Baseplate"]



import logging

from datetime import timedelta
from typing import Any
from typing import Tuple

from zope.interface import implementer

from baseplate.lib.crypto import make_signature
from baseplate.lib.crypto import SignatureError
from baseplate.lib.crypto import validate_signature
from baseplate.lib.secrets import SecretsStore
from baseplate.lib.secrets import VersionedSecret


logger = logging.getLogger(__name__)

try:
    # ICSRFStoragePolicy was not added to Pyramid until version 1.9
    from pyramid.interfaces import ICSRFStoragePolicy  # pylint: disable=no-name-in-module
except ImportError:
    logger.error(
        "baseplate.frameworks.pyramid.csrf requires that you use a version of pyramid >= 1.9"
    )
    raise


def _make_csrf_token_payload(version: int, account_id: str) -> Tuple[str, str]:
    version_str = str(version)
    payload = ".".join([version_str, account_id])
    return version_str, payload


@implementer(ICSRFStoragePolicy)
class TokenCSRFStoragePolicy:
    """ICSRFStoragePolicy implementation for Intranet Services.

    This implementation of Pyramid's ICSRFStoragePolicy interface takes
    advantage of intranet authentication being handled outside individual
    applications and validates CSRF tokens without storing anything at all.
    It works by using a secret value to make HMAC CSRF tokens that are
    scoped to individual users.

    This policy relies on Pyramid's built in Authentication Policies since it
    uses `request.authenticated_userid`.  For a simple, intranet app, you can
    rely on the HTTP "Authenticated-User" header to pass you this value and you
    can configure Pyramid to use that with the following snippet:

        authn_policy = RemoteUserAuthenticationPolicy(environ_key="HTTP_AUTHENTICATED_USER")
        authz_policy = ACLAuthorizationPolicy()
        configurator.set_authentication_policy(authn_policy)
        configurator.set_authorization_policy(authz_policy)

    You can add this form of CSRF protection to your baseplate Pyramid app
    by adding the following code to your `make_wsgi_app` function:

        configurator.set_csrf_storage_policy(TokenCSRFStoragePolicy(
            secrets=secrets,
            secret_path='secret/path/to/csrf-secret',
        ))
        configurator.set_default_csrf_options(require_csrf=True)

    You will also need to pass a new CSRF token to your client each time you
    render a form for them.  You should not re-use CSRF tokens when using this
    StoragePolicy since these tokens expire and are difficult to selectively
    invalidate.

    :param secrets: A SecretsStore that contains the secret you will use to
        sign the CSRF token.
    :param secret_path: The key to the secret in the supplied SecretsStore
    :param param: The name of the parameter to get the CSRF token from on a
        request.  The default is 'csrf_token'.
    :param max_age: The maximum age that the signature portion of the CSRF
        token is valid.  The default value is one hour.
    """

    VERSION = 1

    def __init__(
        self,
        secrets: SecretsStore,
        secret_path: str,
        param: str = "csrf_token",
        max_age: timedelta = timedelta(hours=1),
    ):
        self._secrets = secrets
        self._secret_path = secret_path
        self._param = param
        self._max_age = max_age

    def _get_secret(self) -> VersionedSecret:
        return self._secrets.get_versioned(self._secret_path)

    def new_csrf_token(self, request: Any) -> str:
        """Return a new CSRF token.

        You will need to call `pyramid.csrf.new_csrf_token` to get a new
        CSRF token when rendering forms, you should not re-use CSRF tokens
        when using this StoragePolicy since these tokens expire and are
        difficult to selectively invalidate.
        """
        prefix, payload = _make_csrf_token_payload(
            version=self.VERSION, account_id=request.authenticated_userid
        )
        signature = make_signature(self._get_secret(), payload, self._max_age)
        return ".".join([prefix, signature.decode("utf-8")])

    def get_csrf_token(self, request: Any) -> str:
        """Return the currently active CSRF token from the request params.

        This will not generate a new one if none is supplied like some of
        the default ones in Pyramid do.

        This is called automatically by Pyramid if you have configured it
        to require CSRF.
        """
        return request.params.get(self._param)

    def check_csrf_token(self, request: Any, supplied_token: str) -> bool:
        """Return True if the supplied_token is valid.

        This is called automatically by Pyramid if you have configured it
        to require CSRF.
        """
        try:
            version_str, sep, signature = supplied_token.partition(".")
            token_version = int(version_str)
        except Exception:
            return False

        if sep != ".":
            return False

        if token_version != self.VERSION:
            return False

        _, payload = _make_csrf_token_payload(
            version=token_version, account_id=request.authenticated_userid
        )

        try:
            validate_signature(self._get_secret(), payload, signature.encode())
        except SignatureError:
            return False

        return True



import datetime
import logging
import socket

from typing import Any
from typing import Dict
from typing import Tuple
from typing import Union

from gevent.pool import Pool
from gevent.server import StreamServer
from thrift.protocol.THeaderProtocol import THeaderProtocolFactory
from thrift.Thrift import TProcessor
from thrift.transport.THeaderTransport import THeaderClientType
from thrift.transport.TSocket import TSocket
from thrift.transport.TTransport import TBufferedTransportFactory
from thrift.transport.TTransport import TTransportException

from baseplate.lib import config
from baseplate.server import runtime_monitor


logger = logging.getLogger(__name__)


Address = Union[Tuple[str, int], str]


# pylint: disable=too-many-public-methods
class GeventServer(StreamServer):
    def __init__(self, processor: TProcessor, *args: Any, **kwargs: Any):
        self.processor = processor
        self.transport_factory = TBufferedTransportFactory()
        self.protocol_factory = THeaderProtocolFactory(
            # allow non-headerprotocol clients to talk with us
            allowed_client_types=[
                THeaderClientType.HEADERS,
                THeaderClientType.FRAMED_BINARY,
                THeaderClientType.UNFRAMED_BINARY,
            ]
        )
        super().__init__(*args, **kwargs)

    # pylint: disable=method-hidden,unused-argument
    def handle(self, client_socket: socket.socket, address: Address) -> None:
        client = TSocket()
        client.setHandle(client_socket)

        trans = self.transport_factory.getTransport(client)
        prot = self.protocol_factory.getProtocol(trans)

        try:
            while self.started:
                self.processor.process(prot, prot)
        except TTransportException:
            pass
        finally:
            trans.close()


def make_server(server_config: Dict[str, str], listener: socket.socket, app: Any) -> StreamServer:
    # pylint: disable=maybe-no-member
    cfg = config.parse_config(
        server_config,
        {
            "max_concurrency": config.Optional(config.Integer),
            "stop_timeout": config.Optional(
                config.TimespanWithLegacyFallback, default=datetime.timedelta(seconds=10)
            ),
        },
    )

    if cfg.max_concurrency is not None:
        raise Exception(
            "The max_concurrency setting is not allowed for Thrift servers. See https://github.com/reddit/baseplate.py-upgrader/wiki/v1.2#max_concurrency-is-deprecated."
        )

    pool = Pool()
    server = GeventServer(processor=app, listener=listener, spawn=pool)
    server.stop_timeout = cfg.stop_timeout.total_seconds()

    runtime_monitor.start(server_config, app, pool)
    return server



"""Helpers that integrate common client libraries with Baseplate.py.

This package contains modules which integrate various client libraries with
Baseplate.py's instrumentation. When using these client library integrations,
trace information is passed on and metrics are collected automatically.

"""
from typing import Any
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    import baseplate.lib.metrics


class ContextFactory:
    """An interface for adding stuff to the context object.

    Objects implementing this interface can be passed to
    :py:meth:`~baseplate.Baseplate.add_to_context`. The return value of
    :py:meth:`make_object_for_context` will be added to the
    :py:class:`~baseplate.RequestContext` with the name specified in
    ``add_to_context``.

    """

    def report_runtime_metrics(self, batch: "baseplate.lib.metrics.Client") -> None:
        """Report runtime metrics to the stats system.

        :param batch: A metrics client to report statistics to.

        """

    def make_object_for_context(self, name: str, span: "baseplate.Span") -> Any:
        """Return an object that can be added to the context object.

        :param name: The name assigned to this object on the context.
        :param span: The current span this object is being made for.

        """
        raise NotImplementedError



import contextlib
import inspect
import sys
import time

from math import ceil
from typing import Any
from typing import Callable
from typing import Iterator
from typing import Optional

from prometheus_client import Counter
from prometheus_client import Gauge
from prometheus_client import Histogram
from thrift.protocol.TProtocol import TProtocolException
from thrift.Thrift import TApplicationException
from thrift.Thrift import TException
from thrift.transport.TTransport import TTransportException

from baseplate import Span
from baseplate.clients import ContextFactory
from baseplate.lib import config
from baseplate.lib import metrics
from baseplate.lib.prometheus_metrics import default_latency_buckets
from baseplate.lib.retry import RetryPolicy
from baseplate.lib.thrift_pool import thrift_pool_from_config
from baseplate.lib.thrift_pool import ThriftConnectionPool
from baseplate.thrift.ttypes import Error
from baseplate.thrift.ttypes import ErrorCode

PROM_NAMESPACE = "thrift_client"

PROM_COMMON_LABELS = [
    "thrift_method",
    "thrift_client_name",
]
REQUESTS_TOTAL_LABELS = [
    *PROM_COMMON_LABELS,
    "thrift_success",
    "thrift_exception_type",
    "thrift_baseplate_status",
    "thrift_baseplate_status_code",
]

REQUEST_LATENCY = Histogram(
    f"{PROM_NAMESPACE}_latency_seconds",
    "Latency of thrift client requests",
    [
        *PROM_COMMON_LABELS,
        "thrift_success",
    ],
    buckets=default_latency_buckets,
)

REQUESTS_TOTAL = Counter(
    f"{PROM_NAMESPACE}_requests_total",
    "Total number of outgoing requests",
    REQUESTS_TOTAL_LABELS,
)

ACTIVE_REQUESTS = Gauge(
    f"{PROM_NAMESPACE}_active_requests",
    "number of in-flight requests",
    PROM_COMMON_LABELS,
)


class ThriftClient(config.Parser):
    """Configure a Thrift client.

    This is meant to be used with
    :py:meth:`baseplate.Baseplate.configure_context`.

    See :py:func:`baseplate.lib.thrift_pool.thrift_pool_from_config` for available
    configuration settings.

    :param client_cls: The class object of a Thrift-generated client class,
        e.g. ``YourService.Client``.

    """

    def __init__(self, client_cls: Any, **kwargs: Any):
        self.client_cls = client_cls
        self.kwargs = kwargs

    def parse(self, key_path: str, raw_config: config.RawConfig) -> ContextFactory:
        pool = thrift_pool_from_config(raw_config, prefix=f"{key_path}.", **self.kwargs)
        return ThriftContextFactory(pool, self.client_cls)


class ThriftContextFactory(ContextFactory):
    """Thrift client pool context factory.

    This factory will attach a proxy object with the same interface as your
    thrift client to an attribute on the :py:class:`~baseplate.RequestContext`.
    When a thrift method is called on this proxy object, it will check out a
    connection from the connection pool and execute the RPC, automatically
    recording diagnostic information.

    :param pool: The connection pool.
    :param client_cls: The class object of a Thrift-generated client class,
        e.g. ``YourService.Client``.

    The proxy object has a ``retrying`` method which takes the same parameters
    as :py:meth:`RetryPolicy.new <baseplate.lib.retry.RetryPolicy.new>` and acts as
    a context manager. The context manager returns another proxy object where
    Thrift service method calls will be automatically retried with the
    specified retry policy when transient errors occur::

        with context.my_service.retrying(attempts=3) as svc:
            svc.some_method()

    """

    POOL_PREFIX = "thrift_client_pool"
    POOL_LABELS = ["thrift_pool"]

    max_connections_gauge = Gauge(
        f"{POOL_PREFIX}_max_size",
        "Maximum number of connections in this thrift pool before blocking",
        POOL_LABELS,
    )

    active_connections_gauge = Gauge(
        f"{POOL_PREFIX}_active_connections",
        "Number of connections currently in use in this thrift pool",
        POOL_LABELS,
    )

    def __init__(self, pool: ThriftConnectionPool, client_cls: Any):
        self.pool = pool
        self.client_cls = client_cls
        self.proxy_cls = type(
            "PooledClientProxy",
            (_PooledClientProxy,),
            {
                fn_name: _build_thrift_proxy_method(fn_name)
                for fn_name in _enumerate_service_methods(client_cls)
                if not (fn_name.startswith("__") and fn_name.endswith("__"))
            },
        )

    def report_runtime_metrics(self, batch: metrics.Client) -> None:
        pool_name = self.client_cls.__name__
        self.max_connections_gauge.labels(pool_name).set(self.pool.size)
        self.active_connections_gauge.labels(pool_name).set(self.pool.checkedout)
        batch.gauge("pool.size").replace(self.pool.size)
        batch.gauge("pool.in_use").replace(self.pool.checkedout)
        # it's hard to report "open_and_available" currently because we can't
        # distinguish easily between available connection slots that aren't
        # instantiated and ones that have actual open connections.

    def make_object_for_context(self, name: str, span: Span) -> "_PooledClientProxy":
        return self.proxy_cls(self.client_cls, self.pool, span, name)


def _enumerate_service_methods(client: Any) -> Iterator[str]:
    """Return an iterable of service methods from a generated Iface class."""
    ifaces_found = 0

    # python3 drops the concept of unbound methods, so they're just plain
    # functions and we have to account for that here. see:
    # https://stackoverflow.com/questions/17019949/why-is-there-a-difference-between-inspect-ismethod-and-inspect-isfunction-from-p  # noqa: E501
    def predicate(x: Any) -> bool:
        return inspect.isfunction(x) or inspect.ismethod(x)

    for base_cls in inspect.getmro(client):
        if base_cls.__name__ == "Iface":
            for name, _ in inspect.getmembers(base_cls, predicate):
                yield name
            ifaces_found += 1

    assert ifaces_found > 0, "class is not a thrift client; it has no Iface"


class _PooledClientProxy:
    """A proxy which acts like a thrift client but uses a connection pool."""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        client_cls: Any,
        pool: ThriftConnectionPool,
        server_span: Span,
        namespace: str,
        retry_policy: Optional[RetryPolicy] = None,
    ):
        self.client_cls = client_cls
        self.pool = pool
        self.server_span = server_span
        self.namespace = namespace
        self.retry_policy = retry_policy or RetryPolicy.new(attempts=1)

    @contextlib.contextmanager
    def retrying(self, **policy: Any) -> Iterator["_PooledClientProxy"]:
        yield self.__class__(
            self.client_cls,
            self.pool,
            self.server_span,
            self.namespace,
            retry_policy=RetryPolicy.new(**policy),
        )


def _build_thrift_proxy_method(name: str) -> Callable[..., Any]:
    def _call_thrift_method(self: Any, *args: Any, **kwargs: Any) -> Any:
        trace_name = f"{self.namespace}.{name}"
        last_error = None

        for time_remaining in self.retry_policy:
            try:
                with self.pool.connection() as prot, ACTIVE_REQUESTS.labels(
                    thrift_method=name, thrift_client_name=self.namespace
                ).track_inprogress():
                    start_time = time.perf_counter()

                    span = self.server_span.make_child(trace_name)
                    span.set_tag("slug", self.namespace)

                    client = self.client_cls(prot)
                    method = getattr(client, name)
                    span.set_tag("method", method.__name__)
                    span.start()

                    try:
                        baseplate = span.baseplate
                        if baseplate:
                            service_name = baseplate.service_name
                            if service_name:
                                prot.trans.set_header(b"User-Agent", service_name.encode())

                        prot.trans.set_header(b"Trace", str(span.trace_id).encode())
                        prot.trans.set_header(b"Parent", str(span.parent_id).encode())
                        prot.trans.set_header(b"Span", str(span.id).encode())
                        if span.sampled is not None:
                            sampled = "1" if span.sampled else "0"
                            prot.trans.set_header(b"Sampled", sampled.encode())
                        if span.flags:
                            prot.trans.set_header(b"Flags", str(span.flags).encode())

                        min_timeout = time_remaining
                        if self.pool.timeout:
                            if not min_timeout or self.pool.timeout < min_timeout:
                                min_timeout = self.pool.timeout
                        if min_timeout and min_timeout > 0:
                            # min_timeout is in float seconds, we are converting to int milliseconds
                            # rounding up here.
                            prot.trans.set_header(
                                b"Deadline-Budget", str(int(ceil(min_timeout * 1000))).encode()
                            )

                        try:
                            edge_context = span.context.raw_edge_context
                        except AttributeError:
                            edge_context = None

                        if edge_context:
                            prot.trans.set_header(b"Edge-Request", edge_context)

                        result = method(*args, **kwargs)
                    except TTransportException as exc:
                        # the connection failed for some reason, retry if able
                        span.finish(exc_info=sys.exc_info())
                        last_error = str(exc)
                        if exc.inner is not None:
                            last_error += f" ({exc.inner})"
                        raise  # we need to raise all exceptions so that self.pool.connect() self-heals
                    except (TApplicationException, TProtocolException):
                        # these are subclasses of TException but aren't ones that
                        # should be expected in the protocol. this is an error!
                        span.finish(exc_info=sys.exc_info())
                        raise
                    except Error as exc:
                        # a 5xx error is an unexpected exception but not 5xx are
                        # not.
                        if 500 <= exc.code < 600:
                            span.finish(exc_info=sys.exc_info())
                        else:
                            span.finish()
                        raise
                    except TException:
                        # this is an expected exception, as defined in the IDL
                        span.finish()
                        raise
                    except:  # noqa: E722
                        # something unexpected happened
                        span.finish(exc_info=sys.exc_info())
                        raise
                    else:
                        # a normal result
                        span.finish()
                        return result
                    finally:
                        thrift_success = "true"
                        exception_type = ""
                        baseplate_status = ""
                        baseplate_status_code = ""
                        exc_info = sys.exc_info()
                        if exc_info[0] is not None:
                            thrift_success = "false"
                            exception_type = exc_info[0].__name__
                            current_exc = exc_info[1]
                            try:
                                # We want the following code to execute whenever the
                                # service raises an instance of Baseplate's `Error` class.
                                # Unfortunately, we cannot just rely on `isinstance` to do
                                # what we want here because some services compile
                                # Baseplate's thrift file on their own and import `Error`
                                # from that. When this is done, `isinstance` will always
                                # return `False` since it's technically a different class.
                                # To fix this, we optimistically try to access `code` on
                                # `current_exc` and just catch the `AttributeError` if the
                                # `code` attribute is not present.
                                # Note: if the error code was not originally defined in baseplate, or the
                                # name associated with the error was overriden, this cannot reflect that
                                # we will emit the status code in both cases
                                # but the status will be blank in the first case, and the baseplate name
                                # in the second
                                baseplate_status_code = current_exc.code  # type: ignore
                                baseplate_status = ErrorCode()._VALUES_TO_NAMES.get(current_exc.code, "")  # type: ignore
                            except AttributeError:
                                pass

                        REQUEST_LATENCY.labels(
                            thrift_method=name,
                            thrift_client_name=self.namespace,
                            thrift_success=thrift_success,
                        ).observe(time.perf_counter() - start_time)

                        REQUESTS_TOTAL.labels(
                            thrift_method=name,
                            thrift_client_name=self.namespace,
                            thrift_success=thrift_success,
                            thrift_exception_type=exception_type,
                            thrift_baseplate_status_code=baseplate_status_code,
                            thrift_baseplate_status=baseplate_status,
                        ).inc()

            except TTransportException:
                # swallow exception so we can retry on TTransportException (relies on the for loop)
                continue

        # this only happens if we exhaust the retry policy
        raise TTransportException(
            type=TTransportException.TIMED_OUT,
            message=f"retry policy exhausted while attempting {self.namespace}.{name}, last error was: {last_error}",
        )

    return _call_thrift_method



from pythonjsonlogger import jsonlogger


class CustomJsonFormatter(jsonlogger.JsonFormatter):
    def process_log_record(self, log_record: dict) -> dict:
        log_record["level"] = log_record.pop("levelname", None)
        try:
            # see if there's an integer thread name, that's probably a traceID
            # from the logging observer.
            trace_id = int(log_record["threadName"])

            # unfortunately we have to convert back to string because our log
            # processor can't handle giant integers.
            log_record["traceID"] = str(trace_id)
        except (KeyError, ValueError):
            pass
        return super().process_log_record(log_record)



import queue
import socket
import unittest

from unittest import mock

from thrift.protocol import TBinaryProtocol
from thrift.protocol import THeaderProtocol
from thrift.Thrift import TException
from thrift.transport import THeaderTransport
from thrift.transport import TSocket
from thrift.transport import TTransport

from baseplate.lib import config
from baseplate.lib import thrift_pool
from baseplate.observers.timeout import ServerTimeout


EXAMPLE_ENDPOINT = config.EndpointConfiguration(socket.AF_INET, ("127.0.0.1", 1234))


class MakeTransportTests(unittest.TestCase):
    def test_inet(self):
        endpoint = config.EndpointConfiguration(socket.AF_INET, ("localhost", 1234))
        socket_transport = thrift_pool._make_transport(endpoint)

        self.assertFalse(socket_transport._unix_socket)

    def test_unix(self):
        endpoint = config.EndpointConfiguration(socket.AF_UNIX, "/tmp/socket")
        socket_transport = thrift_pool._make_transport(endpoint)

        self.assertTrue(socket_transport._unix_socket)

    def test_unknown(self):
        endpoint = config.EndpointConfiguration(socket.AF_UNSPEC, None)

        with self.assertRaises(Exception):
            thrift_pool._make_transport(endpoint)


@mock.patch("baseplate.lib.thrift_pool.RetryPolicy.new")
class MaxRetriesRenameTests(unittest.TestCase):
    def test_default_is_3(self, new_retry_policy):
        thrift_pool.ThriftConnectionPool(EXAMPLE_ENDPOINT)
        new_retry_policy.assert_called_with(attempts=3)

    def test_default_through_parser(self, new_retry_policy):
        config = {"example.endpoint": "127.0.0.1:1234"}
        thrift_pool.thrift_pool_from_config(config, prefix="example.")
        new_retry_policy.assert_called_with(attempts=3)

    def test_max_retries_config_error(self, new_retry_policy):
        config = {
            "example.endpoint": "127.0.0.1:1234",
            "example.max_retries": "5",
        }
        with self.assertRaises(Exception):
            thrift_pool.thrift_pool_from_config(config, prefix="example.")

    def test_max_connection_attempts_works(self, new_retry_policy):
        thrift_pool.ThriftConnectionPool(EXAMPLE_ENDPOINT, max_connection_attempts=5)
        new_retry_policy.assert_called_with(attempts=5)

    def test_max_connection_attempts_works_through_config(self, new_retry_policy):
        config = {"example.endpoint": "127.0.0.1:1234", "example.max_connection_attempts": "5"}
        thrift_pool.thrift_pool_from_config(config, prefix="example.")
        new_retry_policy.assert_called_with(attempts=5)


class ThriftFIFOPoolTests(unittest.TestCase):
    def test_fifo_queue_config(self):
        config = {
            "fifo-test.endpoint": "127.0.0.1:1234",
            "fifo-test.fifo_queue": "true",
        }
        pool = thrift_pool.thrift_pool_from_config(config, prefix="fifo-test.")
        self.assertTrue(isinstance(pool.pool, queue.Queue))


class ThriftConnectionPoolTests(unittest.TestCase):
    def setUp(self):
        self.mock_queue = mock.Mock(spec=queue.Queue)
        self.pool = thrift_pool.ThriftConnectionPool(EXAMPLE_ENDPOINT)
        self.pool.pool = self.mock_queue

    def test_pool_empty_timeout(self):
        self.mock_queue.get.side_effect = queue.Empty

        with self.assertRaises(TTransport.TTransportException):
            self.pool._get_from_pool()

    def test_pool_with_framed_protocol_factory(self):
        def framed_protocol_factory(trans):
            trans = TTransport.TFramedTransport(trans)
            return TBinaryProtocol.TBinaryProtocol(trans)

        framed_pool = thrift_pool.ThriftConnectionPool(
            EXAMPLE_ENDPOINT, protocol_factory=framed_protocol_factory
        )
        trans = thrift_pool._make_transport(EXAMPLE_ENDPOINT)
        prot = framed_pool.protocol_factory(trans)

        self.assertTrue(isinstance(prot, TBinaryProtocol.TBinaryProtocol))
        self.assertTrue(isinstance(prot.trans, TTransport.TFramedTransport))

    @mock.patch("time.time")
    def test_pool_has_valid_connection(self, mock_time):
        mock_time.return_value = 123
        mock_prot = mock.Mock(spec=THeaderProtocol.THeaderProtocol)
        mock_prot.baseplate_birthdate = 122
        self.mock_queue.get.return_value = mock_prot

        prot = self.pool._get_from_pool()

        self.assertEqual(prot, mock_prot)

    @mock.patch("baseplate.lib.thrift_pool._make_transport")
    @mock.patch("time.time")
    def test_pool_closes_stale_connection(self, mock_time, mock_make_transport):
        stale_prot = mock.Mock(spec=THeaderProtocol.THeaderProtocol)
        stale_prot.trans = mock.Mock(spec=TSocket.TSocket)

        stale_prot.baseplate_birthdate = 10
        mock_time.return_value = 200
        self.mock_queue.get.return_value = stale_prot

        fresh_trans = mock.Mock(spec=TSocket.TSocket)
        fresh_trans.protocol_id = THeaderTransport.THeaderSubprotocolID.BINARY
        mock_make_transport.return_value = fresh_trans

        with self.pool.connection() as prot:
            pass

        self.assertTrue(stale_prot.trans.close.called)
        self.assertEqual(prot.trans._transport, fresh_trans)

    @mock.patch("baseplate.lib.thrift_pool._make_transport")
    @mock.patch("time.time")
    def test_pool_replaces_closed_connection(self, mock_time, mock_make_transport):
        stale_prot = mock.Mock(spec=THeaderProtocol.THeaderProtocol)
        stale_prot.trans = mock.Mock(spec=TSocket.TSocket)
        stale_prot.trans.isOpen.return_value = False
        stale_prot.baseplate_birthdate = 199
        self.mock_queue.get.return_value = stale_prot
        mock_time.return_value = 200

        fresh_trans = mock.Mock(spec=TSocket.TSocket)
        fresh_trans.protocol_id = THeaderTransport.THeaderSubprotocolID.BINARY
        mock_make_transport.return_value = fresh_trans

        with self.pool.connection() as prot:
            pass

        self.assertTrue(stale_prot.trans.close.called)
        self.assertEqual(prot.trans._transport, fresh_trans)

    @mock.patch("baseplate.lib.thrift_pool._make_transport")
    @mock.patch("time.time")
    def test_retry_on_failed_connect(self, mock_time, mock_make_transport):
        self.mock_queue.get.return_value = None
        mock_time.return_value = 200

        broken_trans = mock.Mock(spec=TSocket.TSocket)
        broken_trans.protocol_id = THeaderTransport.THeaderSubprotocolID.BINARY
        broken_trans.open.side_effect = TTransport.TTransportException

        ok_trans = mock.Mock(spec=TSocket.TSocket)
        ok_trans.protocol_id = THeaderTransport.THeaderSubprotocolID.BINARY

        mock_make_transport.side_effect = [broken_trans, ok_trans]

        prot = self.pool._create_connection()

        self.assertEqual(prot.trans._transport, ok_trans)
        self.assertEqual(ok_trans.open.call_count, 1)

    @mock.patch("baseplate.lib.thrift_pool._make_transport")
    @mock.patch("time.time")
    def test_max_retry_on_connect(self, mock_time, mock_make_transport):
        self.mock_queue.get.return_value = None
        mock_time.return_value = 200

        broken_trans = mock.Mock(spec=TSocket.TSocket)
        broken_trans.protocol_id = THeaderTransport.THeaderSubprotocolID.BINARY
        broken_trans.open.side_effect = TTransport.TTransportException

        mock_make_transport.side_effect = [broken_trans] * 3

        with self.assertRaises(TTransport.TTransportException):
            self.pool._create_connection()

    def test_release_open(self):
        mock_prot = mock.Mock(spec=THeaderProtocol.THeaderProtocol)
        mock_prot.trans = mock.Mock(spec=TSocket.TSocket)
        mock_prot.trans.isOpen.return_value = True

        self.pool._release(mock_prot)

        self.assertEqual(self.mock_queue.put.call_count, 1)
        self.assertEqual(self.mock_queue.put.call_args, mock.call(mock_prot))

    def test_release_closed(self):
        mock_prot = mock.Mock(spec=THeaderProtocol.THeaderProtocol)
        mock_prot.trans = mock.Mock(spec=TSocket.TSocket)
        mock_prot.trans.isOpen.return_value = False

        self.pool._release(mock_prot)

        self.assertEqual(self.mock_queue.put.call_count, 1)
        self.assertEqual(self.mock_queue.put.call_args, mock.call(None))

    @mock.patch("time.time")
    def test_context_normal(self, mock_time):
        mock_time.return_value = 123
        mock_prot = mock.Mock(spec=THeaderProtocol.THeaderProtocol)
        mock_prot.baseplate_birthdate = 122
        mock_prot.trans = mock.Mock(spec=TSocket.TSocket)
        mock_prot.trans.isOpen.return_value = True
        self.mock_queue.get.return_value = mock_prot

        with self.pool.connection() as prot:
            pass

        self.assertEqual(prot, mock_prot)
        self.assertEqual(self.mock_queue.get.call_count, 1)
        self.assertEqual(self.mock_queue.put.call_count, 1)
        self.assertEqual(self.mock_queue.put.call_args, mock.call(mock_prot))

    @mock.patch("time.time")
    def test_context_thrift_exception(self, mock_time):
        mock_time.return_value = 123
        mock_prot = mock.Mock(spec=THeaderProtocol.THeaderProtocol)
        mock_prot.baseplate_birthdate = 122
        mock_prot.trans = mock.Mock(spec=TSocket.TSocket)
        mock_prot.trans.isOpen.return_value = True
        self.mock_queue.get.return_value = mock_prot

        with self.assertRaises(TException):
            with self.pool.connection() as prot:
                mock_prot.trans.isOpen.return_value = False
                raise TTransport.TTransportException

        self.assertEqual(prot, mock_prot)
        self.assertEqual(mock_prot.trans.close.call_count, 1)
        self.assertEqual(self.mock_queue.get.call_count, 1)
        self.assertEqual(self.mock_queue.put.call_count, 1)
        self.assertEqual(self.mock_queue.put.call_args, mock.call(None))

    @mock.patch("time.time")
    def test_context_non_thrift_exception(self, mock_time):
        mock_time.return_value = 123
        mock_prot = mock.Mock(spec=THeaderProtocol.THeaderProtocol)
        mock_prot.baseplate_birthdate = 122
        mock_prot.trans = mock.Mock(spec=TSocket.TSocket)
        mock_prot.trans.isOpen.return_value = True
        self.mock_queue.get.return_value = mock_prot

        with self.assertRaises(Exception):
            with self.pool.connection() as prot:
                raise Exception

        self.assertEqual(prot, mock_prot)
        self.assertEqual(self.mock_queue.get.call_count, 1)
        self.assertEqual(self.mock_queue.put.call_count, 1)
        self.assertEqual(self.mock_queue.put.call_args, mock.call(mock_prot))

    @mock.patch("baseplate.lib.thrift_pool._make_transport")
    @mock.patch("time.time")
    def test_context_server_timeout(self, mock_time, mock_make_transport):
        mock_time.return_value = 123

        self.mock_queue.get.return_value = None

        fresh_trans = mock.Mock(spec=TSocket.TSocket)
        fresh_trans.protocol_id = THeaderTransport.THeaderSubprotocolID.BINARY
        fresh_trans.open.side_effect = ServerTimeout("span", 10.0, False)
        mock_make_transport.return_value = fresh_trans

        with self.assertRaises(ServerTimeout):
            with self.pool.connection() as _:
                pass

        self.assertEqual(self.mock_queue.get.call_count, 1)
        self.assertEqual(self.mock_queue.put.call_count, 1)
        self.assertEqual(self.mock_queue.put.call_args, mock.call(None))

    @mock.patch("baseplate.lib.thrift_pool._make_transport")
    @mock.patch("time.time")
    def test_context_t_transport_exception(self, mock_time, mock_make_transport):
        mock_time.return_value = 123

        self.mock_queue.get.return_value = None

        broken_trans = mock.Mock(spec=TSocket.TSocket)
        broken_trans.protocol_id = THeaderTransport.THeaderSubprotocolID.BINARY
        broken_trans.open.side_effect = TTransport.TTransportException
        mock_make_transport.return_value = broken_trans

        with self.assertRaises(TTransport.TTransportException):
            with self.pool.connection() as _:
                pass

        self.assertEqual(self.mock_queue.get.call_count, 1)
        self.assertEqual(self.mock_queue.put.call_count, 1)
        self.assertEqual(self.mock_queue.put.call_args, mock.call(None))

    def test_pool_checkout_exception(self):
        class PatchedLifoQueue(queue.LifoQueue):
            def get(self, *args, **kwargs):
                raise Exception

        pool = thrift_pool.ThriftConnectionPool(EXAMPLE_ENDPOINT, queue_cls=PatchedLifoQueue)

        with self.assertRaises(Exception):
            with pool.connection() as _:
                pass

        self.assertEqual(0, pool.checkedout)



"""A Gevent-friendly POSIX message queue."""
import select

from typing import Optional

import posix_ipc

from baseplate.lib.retry import RetryPolicy


class MessageQueueError(Exception):
    """Base exception for message queue related errors."""


class TimedOutError(MessageQueueError):
    """Raised when a message queue operation times out."""

    def __init__(self) -> None:
        super().__init__("Timed out waiting for the message queue.")


# this wrapper-exception is here just to give the user a bit more of an idea
# how to fix the error should they run into it since the base error message
# is rather opaque.
class InvalidParametersError(ValueError):
    def __init__(self, inner: Exception):
        super().__init__(f"{inner} (check fs.mqueue.{{msg_max,msgsize_max}} sysctls?)")


# this wrapper-exception is here just to give the user a bit more of an idea
# how to fix the error should they run into it since the base error message
# is rather opaque.
class MessageQueueOSError(OSError):
    def __init__(self, inner: Exception):
        super().__init__(f"{inner} (check `ulimit -q`?)")


class MessageQueue:
    """A Gevent-friendly (but not required) inter process message queue.

    ``name`` should be a string of up to 255 characters consisting of an
    initial slash, followed by one or more characters, none of which are
    slashes.

    Note: This relies on POSIX message queues being available and
    select(2)-able like other file descriptors. Not all operating systems
    support this.

    """

    def __init__(self, name: str, max_messages: int, max_message_size: int):
        try:
            self.queue = posix_ipc.MessageQueue(
                name,
                flags=posix_ipc.O_CREAT,
                mode=0o0644,
                max_messages=max_messages,
                max_message_size=max_message_size,
            )
        except ValueError as exc:
            raise InvalidParametersError(exc)
        except OSError as exc:
            raise MessageQueueOSError(exc)
        self.queue.block = False

    def get(self, timeout: Optional[float] = None) -> bytes:
        """Read a message from the queue.

        :param timeout: If the queue is empty, the call will block up to
            ``timeout`` seconds or forever if ``None``.
        :raises: :py:exc:`TimedOutError` The queue was empty for the allowed
            duration of the call.

        """
        for time_remaining in RetryPolicy.new(budget=timeout):
            try:
                message, _ = self.queue.receive()
                return message
            except posix_ipc.SignalError:  # pragma: nocover
                continue  # interrupted, just try again
            except posix_ipc.BusyError:
                select.select([self.queue.mqd], [], [], time_remaining)

        raise TimedOutError

    def put(self, message: bytes, timeout: Optional[float] = None) -> None:
        """Add a message to the queue.

        :param timeout: If the queue is full, the call will block up to
            ``timeout`` seconds or forever if ``None``.
        :raises: :py:exc:`TimedOutError` The queue was full for the allowed
            duration of the call.

        """
        for time_remaining in RetryPolicy.new(budget=timeout):
            try:
                return self.queue.send(message=message)
            except posix_ipc.SignalError:  # pragma: nocover
                continue  # interrupted, just try again
            except posix_ipc.BusyError:
                select.select([], [self.queue.mqd], [], time_remaining)

        raise TimedOutError

    def unlink(self) -> None:
        """Remove the queue from the system.

        The queue will not leave until the last active user closes it.

        """
        self.queue.unlink()

    def close(self) -> None:
        """Close the queue, freeing related resources.

        This must be called explicitly if queues are created/destroyed on the
        fly. It is not automatically called when the object is reclaimed by
        Python.

        """
        self.queue.close()


def queue_tool() -> None:
    import argparse
    import sys

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--max-messages",
        type=int,
        default=10,
        help="if creating the queue, what to set the maximum queue length to",
    )
    parser.add_argument(
        "--max-message-size",
        type=int,
        default=8096,
        help="if creating the queue, what to set the maximum message size to",
    )
    parser.add_argument("queue_name", help="the name of the queue to consume")

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "--create",
        action="store_const",
        dest="mode",
        const="create",
        help="create the named queue if it doesn't exist and exit",
    )
    group.add_argument(
        "--read",
        action="store_const",
        dest="mode",
        const="read",
        help="read, log, and discard messages from the named queue",
    )
    group.add_argument(
        "--write",
        action="store_const",
        dest="mode",
        const="write",
        help="read messages from stdin and write them to the named queue",
    )

    args = parser.parse_args()

    queue = MessageQueue(args.queue_name, args.max_messages, args.max_message_size)

    if args.mode == "read":
        while True:
            item = queue.get()
            print(item.decode())
    elif args.mode == "write":
        for line in sys.stdin:
            queue.put(line.rstrip("\n").encode())


if __name__ == "__main__":
    queue_tool()
